# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""A model of the geodetic earth and intersecting paths/spheres."""

from math import sqrt

from upstage_des.data_types import GeodeticLocation
from upstage_des.geography import (
    INTERSECTION_LOCATION_CALLABLE,
    CrossingCondition,
    Spherical,
)
from upstage_des.motion.great_circle_calcs import get_dist_rad, get_great_circle_points
from upstage_des.units import unit_convert


def _to_tuple(loc: GeodeticLocation) -> tuple[float, float, float]:
    """Convert a Geodetic location to a tuple.

    Properly corrects for meters being expected.

    Args:
        loc (GeodeticLocation): Lat/Lon/Alt location

    Returns:
        tuple[float, float, float]: Data as tuple
    """
    _l = loc.to_degrees()
    units: str = loc.stage.altitude_units
    alt = unit_convert(_l.alt, units, "m")
    return (_l.lat, _l.lon, alt)


def subdivide_intersection(
    start: GeodeticLocation,
    finish: GeodeticLocation,
    speed: float,
    sensor_location: GeodeticLocation,
    sensor_radius: float,
) -> tuple[list[GeodeticLocation], list[float], list[str], float]:
    """Numerical intersection calculation.

    Requires:
        UP.add_stage_variable("intersection_model", INTERSECTION_LOCATION_CALLABLE)
        UP.add_stage_variable("distance_units", ...)

    Args:
        start (GeodeticLocation): Path start
        finish (GeodeticLocation): Path end
        speed (float): Speed on path
        sensor_location (GeodeticLocation): Location of a sensor
        sensor_radius (float): Radius of sensor line of sight.

    Returns:
        tuple[list[GeodeticLocation], list[float], list[str], float]: intersections,
        times, types, path_time
    """
    STAGE = start.stage
    alt_units: str = STAGE.altitude_units
    dist_units: str = STAGE.distance_units
    intersection: INTERSECTION_LOCATION_CALLABLE = STAGE.intersection_model
    path_dist = finish - start
    path_time = path_dist / speed
    intersect_locs = intersection(
        _to_tuple(start),
        _to_tuple(finish),
        _to_tuple(sensor_location),
        sensor_radius,
        dist_units,
        STAGE.stage_model,
        9260,
        [10, 20],
    )

    if not intersect_locs:
        return [], [], ["Bad", "Bad"], 0.0

    # convert the data to a more useful format
    intersections: list[GeodeticLocation] = []
    times: list[float] = []
    types: list[str] = []
    condition: CrossingCondition
    for condition in intersect_locs:
        lat, lon, alt = condition.begin
        alt = unit_convert(alt, "m", alt_units)
        the_loc = GeodeticLocation(lat, lon, alt)
        dist_from_start = the_loc - start
        time_from_start = dist_from_start / speed
        intersections.append(the_loc)
        times.append(time_from_start)
        types.append(condition.kind)

    return intersections, times, types, path_time


def analytical_intersection(
    start: GeodeticLocation,
    finish: GeodeticLocation,
    speed: float,
    sensor_location: GeodeticLocation,
    sensor_radius: float,
) -> tuple[list[GeodeticLocation], list[float], list[str], float]:
    """Calculate the intersection of a great circle.

    The circle is defined by start & finish and the sphere defined by
    sensor_location and sensor_radius.

    This function mimics the above `subdivide_intersection`, but uses analytical
    equations that run much faster.

    Requires:
        UP.add_stage_variable("distance_units", ...)
        UP.add_stage_variable("altitude_units", ...)

    Args:
        start (GeodeticLocation): the start location of the mover
        finish (GeodeticLocation): the finish location of the mover
        speed (float): the speed of the mover (in STAGE units)
        sensor_location (GeodeticLocation): the location of the sensor
        sensor_radius (float): the radius of the sensor (in STAGE units)

    Returns:
        tuple[list[GeodeticLocation], list[float], list[str], float]:
        intersections, times, types, path_time
    """
    STAGE = start.stage
    dist_units: str = STAGE.distance_units
    altitude_units: str = STAGE.altitude_units
    # convert some units
    earth_rad = unit_convert(Spherical.EARTH_RADIUS, "m", dist_units)
    start_rad = start.to_radians()
    finish_rad = finish.to_radians()
    sensor_location_rad = sensor_location.to_radians()

    # modify the sensor's radius to account for the altitude of the mover
    average_path_height = (start.alt + finish.alt) / 2.0
    average_path_height_dist_units = unit_convert(average_path_height, altitude_units, dist_units)
    adjusted_sensor_radius = sqrt(sensor_radius**2 - average_path_height_dist_units**2)
    sensor_radius_rad = adjusted_sensor_radius / earth_rad

    # Note: some calcs may be doubled here, further speed up potentially by optimizing this
    point_results = get_great_circle_points(
        start_rad, finish_rad, sensor_location_rad, sensor_radius_rad
    )
    dist_start = get_dist_rad(sensor_location_rad, start_rad)
    dist_finish = get_dist_rad(sensor_location_rad, finish_rad)
    path_dist = get_dist_rad(start_rad, finish_rad)
    path_time = (path_dist * earth_rad) / speed

    # no intersection
    if point_results is None:
        # return matches subdivide_intersection results
        return [], [], ["Bad", "Bad"], -1.0

    points, distances = point_results
    # get points, times, and types
    intersections = []
    times = []
    types = []
    alt_change_per_dist = (finish.alt - start.alt) / path_dist  # altitude change per radian

    # check for start in the sensor range
    if dist_start < sensor_radius_rad:
        types.append("START_INSIDE")
        intersections.append(start)
        times.append(0.0)
    else:
        types.append("ENTER")

        # estimate intersection altitude assuming linear change
        alt_1 = start.alt + (distances[0] / path_dist) * alt_change_per_dist
        # adjust distance to account for average altitude from start to first intersection
        d1_m = distances[0] * (
            earth_rad + unit_convert(0.5 * (start.alt + alt_1), altitude_units, dist_units)
        )

        intersections.append(
            GeodeticLocation(
                *points[0],
                alt=alt_1,
                in_radians=True,
            ).to_degrees()
        )
        times.append(d1_m / speed)

    # check for end in the sensor range
    if dist_finish < sensor_radius_rad:
        types.append("END_INSIDE")
        intersections.append(finish)
        d_end = path_dist * earth_rad
        times.append(d_end / speed)
    else:
        types.append("EXIT")

        # estimate intersection altitude assuming linear change
        alt_2 = start.alt + (distances[1] / path_dist) * alt_change_per_dist
        # adjust distance to account for average altitude from start to first intersection
        d2_m = distances[1] * (
            earth_rad + unit_convert(0.5 * (start.alt + alt_2), altitude_units, dist_units)
        )

        intersections.append(GeodeticLocation(*points[1], alt=alt_2, in_radians=True).to_degrees())
        times.append(d2_m / speed)

    return intersections, times, types, path_time



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from collections.abc import Sequence

import pytest
from simpy import Environment
from simpy import Resource as sp_resource
from simpy import Store as sp_store

from upstage_des.api import (
    Actor,
    Any,
    CartesianLocationChangingState,
    DecisionTask,
    EnvironmentContext,
    Event,
    Get,
    InterruptStates,
    LinearChangingState,
    Put,
    ResourceHold,
    State,
    Task,
    TaskLinks,
    TaskNetworkFactory,
    Wait,
    add_stage_variable,
)
from upstage_des.data_types import CartesianLocation, Location
from upstage_des.task import process
from upstage_des.type_help import SIMPY_GEN, TASK_GEN


class Base:
    def __init__(
        self,
        env: Environment,
        name: str,
        x: float,
        y: float,
        num_runways: int = 1,
        parking_max: int = 10,
    ) -> None:
        self.env = env
        self.name = name
        self.location = CartesianLocation(x=x, y=y)
        self.runway = sp_resource(self.env, capacity=num_runways)
        self.maintenance_queue = sp_store(self.env)
        self.parking = sp_store(self.env, parking_max)
        self.parking_tokens = sp_store(self.env, parking_max)
        self.parking_tokens.items = [(self, self.parking_tokens, i) for i in range(parking_max)]
        self.operational = True

        # yes, this is weird
        self.location = CartesianLocation(x, y)

        self.parking_max = parking_max
        self.curr_parked: int = 0
        self.claimed_parking: int = 0
        self._parking_claims: list[Any] = []
        self._parked: list[Any] = []

    def claim_parking(self, plane: Any) -> None:
        self._parking_claims.append(plane)
        self.claimed_parking += 1

    def has_parking(self, plane: Any) -> int:
        return self.curr_parked + self.claimed_parking < self.parking_max

    @process
    def run_maintenance(self) -> SIMPY_GEN:
        while True:
            plane = yield Get(self.maintenance_queue).as_event()
            yield Wait(1.6).as_event()
            mx_wait = plane.get_knowledge("mx_wait")
            # alter the plane's code
            plane.code = 0
            mx_wait.succeed()

    def __repr__(self) -> str:
        return f"{self.name}:{super().__repr__()}"


class World:
    """A helper class for data storage and environment analysis"""

    def __init__(self, bases: Sequence[Base]) -> None:
        self.bases = bases

    def nearest_base(self, loc: CartesianLocation) -> Base:
        b = min(self.bases, key=lambda x: x.location - loc)
        return b

    def bases_by_location(self, loc: CartesianLocation) -> Base:
        x, y = (loc.x, loc.y)
        return [b for b in self.bases if b.location.x == x and b.location.y == y][0]


class Aircraft(Actor):
    base = State[Base | None](default=None)
    location = CartesianLocationChangingState(recording=True)
    speed = State[float]()
    landing_time = State[float]()
    takeoff_time = State[float]()
    code = State[int]()
    fuel = LinearChangingState(recording=True)
    fuel_burn = State[float]()
    parking_token = State[int]()
    parking_spot = State[int]()
    command_data = State[Any]()
    world = State[World]()

    def calculate_bingo(self, max_time: float = float("inf")) -> float:
        # get the farthest base
        farthest: Base = max(self.world.bases, key=lambda x: self.location - x.location)
        dist = self.location - farthest.location
        time = dist / self.speed
        fuel_needed = self.fuel_burn * time
        time_until_bingo = (self.fuel - fuel_needed) / self.fuel_burn
        return min(time_until_bingo, max_time)


class CodeFour(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        wait = Event(rehearsal_time_to_complete=float("inf"))
        yield wait


class GroundWait(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        wait = Event(rehearsal_time_to_complete=0.0)
        yield wait


class GroundTakeoffWait(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        destination = self.get_actor_knowledge(actor, "destination")
        if not isinstance(destination, Location):
            destination = destination.location
        arrival_time = self.get_actor_knowledge(actor, "arrival time")
        if arrival_time is None:
            wait_time = 0
        else:
            distance = destination - actor.location
            flight_time = distance / actor.speed
            wait_time = arrival_time - flight_time
        yield Wait(wait_time)


class Takeoff(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        base = actor.base
        assert base is not None
        runway_request = ResourceHold(base.runway)
        yield runway_request
        takeoff_time = Wait(actor.takeoff_time)
        yield takeoff_time
        yield runway_request
        # HOW WOULD THIS WORK IN TRIAL MODE?
        # what if it's a parking spot token? A token state?
        parking_event = Put(base.parking_tokens, actor.parking_spot)
        yield parking_event
        actor.parking_spot = -1
        # set our current location as above the base
        actor.base = None

    def on_interrupt(self, *, actor: Aircraft, cause: str) -> InterruptStates:
        director = self.get_actor_knowledge(actor, "director")
        if director is not None:
            director.report_failure(actor)
        return self.INTERRUPT.END


class Fly(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        destination = self.get_actor_knowledge(actor, "destination")
        if not isinstance(destination, Location):
            destination = destination.location
        actor.activate_state(
            state="location",
            task=self,
            speed=actor.speed,
            waypoints=[destination],
        )
        actor.activate_state(
            state="fuel",
            task=self,
            rate=-actor.fuel_burn,
        )

        # assume that locations can do this
        distance = destination - actor.location
        time = distance / actor.speed
        fly_wait = Wait(time)

        yield fly_wait

        actor.deactivate_all_states(task=self)
        intent = self.get_actor_knowledge(actor, "intent")
        next_task = self.get_actor_next_task(actor)

        # if our intent is to land, make sure next task is a landing check
        if intent == "land" and next_task != "LandingCheck":
            self.clear_actor_task_queue(actor)
            self.set_actor_task_queue(
                actor,
                [
                    "LandingCheck",
                ],
            )

    def on_interrupt(self, *, actor: Aircraft, cause: str) -> InterruptStates:
        # For testing a restart, this flying is fine, since there is only one
        # final destination.
        if cause == "restart":
            actor.speed = 0.5
            return self.INTERRUPT.RESTART
        else:
            return self.INTERRUPT.END


class Loiter(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        # calculate bingo
        time = actor.calculate_bingo()
        loiter_wait = Wait(time)
        actor.activate_state(
            state="fuel",
            task=self,
            fuel_burn_rate=-actor.fuel_burn,
        )
        yield loiter_wait
        actor.deactivate_state(
            state="fuel",
            task=self,
        )


class Mission(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        # tell the mission folks you are here
        # they'll give you an event to watch to leave
        commander = self.get_actor_knowledge(actor, "commander", must_exist=True)
        leave_event = commander.arrival(actor)
        # set up an alternate bingo leave event
        time = actor.calculate_bingo()
        bingo_wait = Wait(time)

        stop_mission = Any(leave_event, bingo_wait)
        actor.activate_state(
            state="fuel",
            task=self,
            fuel_burn_rate=-actor.fuel_burn,
        )
        yield stop_mission
        actor.deactivate_all_states(task=self)


class LandingLocationSelection(DecisionTask):
    def rehearse_decision(self, *, actor: Aircraft) -> None:
        base = actor.stage.world.bases[0]
        self.set_actor_knowledge(actor, "destination", base)
        self.set_actor_knowledge(actor, "intent", "land")

    def make_decision(self, *, actor: Aircraft) -> None:
        # These kinds of cognitive tasks must be zero-time (no yields!)
        landable_bases = [b for b in actor.stage.world.bases if b.has_parking(actor)]
        base = landable_bases[0]

        self.set_actor_knowledge(actor, "destination", base)
        self.set_actor_knowledge(actor, "intent", "land")


class LandingLocationPrep(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        base = self.get_actor_knowledge(actor, "destination")
        token_event = Get(base.parking_tokens)
        parking_token = yield token_event
        self.set_actor_knowledge(actor, "parking_token", parking_token)


class LandingCheck(DecisionTask):
    """Task for checking the ability to land at the base an actor is above.

    If landing is available, continue in the task network. Otherwise, reselect a base.
    """

    return_task_list = [
        "LandingLocationSelection",
        "LandingLocationPrep",
        "Fly",
        "LandingCheck",
        "Land",
        "MaintenanceWait",
    ]

    def rehearse_decision(self, *, actor: Aircraft) -> None:
        return None

    def make_decision(self, *, actor: Aircraft) -> None:
        # get base from the actor's destination
        base = self.get_actor_knowledge(actor, "destination")
        # assert that the base can be landed at
        if not base.operational:
            # clear the queue
            self.clear_actor_task_queue(actor)
            self.clear_actor_knowledge(actor, "destination")
            self.clear_actor_knowledge(actor, "intent")
            # set up for a task network path that gets a new place to land
            self.set_actor_task_queue(actor, self.return_task_list)
        else:
            # check that we are landing
            msg = f"Actor {actor} not landing after check"
            assert self.get_actor_next_task(actor) == "Land", msg


class Land(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        base = self.get_actor_knowledge(actor, "destination")
        self.clear_actor_knowledge(actor, "destination")
        runway_request = ResourceHold(base.runway)

        self.set_marker("pre-runway")
        yield runway_request
        landing_time = Wait(actor.landing_time)

        self.set_marker("during landing")
        yield landing_time

        self.set_marker("post-landing", self.INTERRUPT.IGNORE)
        yield runway_request
        self.clear_marker()

        self.set_actor_knowledge(actor, "base", base)
        # just to help with a 'clear actor from all stores' need?
        put_event = Put(base.parking, actor)
        self.set_marker("get parking", self.INTERRUPT.IGNORE)
        yield put_event

        self.set_marker("post-parking", interrupt_action=self.INTERRUPT.IGNORE)
        parking_token = self.get_actor_knowledge(actor, "parking_token")
        put_event = Put(base.parking_tokens, parking_token)
        yield put_event

        self.clear_actor_knowledge(actor, "parking_token")
        self.clear_actor_knowledge(actor, "intent")

    def on_interrupt(self, *, actor: Aircraft, cause: str) -> InterruptStates:
        # if interrupted, find a new place to land
        # figure out where we were in the task based on the marker
        marker = self.get_marker()
        self.get_marker_time()
        if marker in [
            "pre-runway",
        ]:
            # We are done with this base
            # put the parking token back
            parking_token = self.get_actor_knowledge(actor, "parking_token")
            if parking_token:
                store = parking_token[1]
                Put(store, parking_token)
                self.clear_actor_knowledge(actor, "parking_token")
            # set the task network to try landing again
            return_task_list = [
                "LandingLocationSelection",
                "LandingLocationPrep",
                "Fly",
                "LandingCheck",
                "Land",
                "MaintenanceWait",
            ]
            self.clear_actor_task_queue(actor)
            self.set_actor_task_queue(actor, return_task_list)
            return self.INTERRUPT.END
        elif marker in [
            "during landing",
        ]:
            # continue on if the cause is benign
            if cause == "Code 4":
                # clear the knowledge and task queue
                parking_token = self.get_actor_knowledge(actor, "parking_token")
                if parking_token:
                    store = parking_token[1]
                    Put(store, parking_token)
                    self.clear_actor_knowledge(actor, "parking_token")
                self.clear_actor_task_queue(actor)
                self.set_actor_task_queue(actor, ["Code4"])
                return self.INTERRUPT.END
            else:
                return self.INTERRUPT.IGNORE
        else:
            # other markers mean that the landing was safe so no
            # need to do anything
            raise ValueError("Shouldn't be here")


class MaintenanceWait(Task):
    def task(self, *, actor: Aircraft) -> TASK_GEN:
        base = self.get_actor_knowledge(actor, "base")
        maintenance_put = Put(base.maintenance_queue, actor)
        yield maintenance_put
        mx_wait = Event(rehearsal_time_to_complete=2.0)
        self.set_actor_knowledge(actor, "mx_wait", mx_wait)
        yield mx_wait
        self.clear_actor_knowledge(actor, "mx_wait")

    def on_interrupt(self, *, actor: Actor, cause: Any) -> InterruptStates:
        return InterruptStates.IGNORE


BASE_LOCATIONS = [
    (10.023, 4.63),
    (2.409, 7.279),
    (0.529, 11.004),
    (6.468, 17.153),
    (5.802, 17.215),
]

task_classes = {
    "GroundWait": GroundWait,
    "GroundTakeoffWait": GroundTakeoffWait,
    "Takeoff": Takeoff,
    "Fly": Fly,
    "Loiter": Loiter,
    "Mission": Mission,
    "LandingLocationSelection": LandingLocationSelection,
    "LandingLocationPrep": LandingLocationPrep,
    "LandingCheck": LandingCheck,
    "Land": Land,
    "MaintenanceWait": MaintenanceWait,
    "Code4": CodeFour,
}
_task_links = {
    "GroundWait": [
        "Takeoff",
        "Code4",
    ],
    "Takeoff": [
        "Fly",
        "Loiter",
        "LandingLocationSelection",
        "Code4",
    ],
    "Loiter": [
        "Fly",
        "Mission",
        "LandingLocationSelection",
    ],
    "Fly": [
        "Loiter",
        "Mission",
        "LandingLocationSelection",
        "LandingCheck",
        "Fly",
    ],
    "Mission": [
        "Fly",
        "Loiter",
        "LandingLocationSelection",
    ],
    "LandingLocationSelection": [
        "Fly",
        "Loiter",
        "LandingLocationPrep",
    ],
    "LandingLocationPrep": [
        "Fly",
        "Loiter",
        "LandingCheck",
        "LandingLocationSelection",
    ],
    "LandingCheck": [
        "Land",
        "Loiter",
        "Fly",
        "LandingLocationSelection",
    ],
    "Land": [
        "MaintenanceWait",
        "Loiter",
        "Code4",
    ],
    "MaintenanceWait": [
        "GroundWait",
        "Code4",
    ],
    "Code4": [],
}
# quick fix for new task network link style
task_links: dict[str, TaskLinks] = {}
for k, v in _task_links.items():
    new = TaskLinks(default=v[0] if v else None, allowed=v)
    task_links[k] = new


def _build_test(env: Environment) -> Aircraft:
    bases = []
    for i in range(len(BASE_LOCATIONS)):
        x, y = BASE_LOCATIONS[i]
        b = Base(env, f"Base {i}", x, y)
        bases.append(b)
        b.run_maintenance()

    world = World(bases)
    add_stage_variable("world", world)

    p = Aircraft(
        name="my plane",
        base=None,
        location=CartesianLocation(0, 0),
        speed=12,
        landing_time=5 / 60,
        takeoff_time=10 / 60,
        code=2,
        fuel=100,
        fuel_burn=15,
        parking_token=None,
        parking_spot=None,
        command_data=None,
        world=world,
        debug_log=True,
    )

    return p


def test_plane_bingo() -> None:
    with EnvironmentContext() as env:
        p = _build_test(env)
        bingo_hours = 5.14
        bingo_result = p.calculate_bingo()
        assert pytest.approx(bingo_result, abs=0.01) == bingo_hours


def test_creating_network() -> None:
    with EnvironmentContext():
        task_fact = TaskNetworkFactory(
            "plane_net",
            task_classes,
            task_links,
        )
        _ = task_fact.make_network()


def test_rehearsing_network() -> None:
    with EnvironmentContext() as env:
        actor = _build_test(env)
        task_fact = TaskNetworkFactory(
            "plane_net",
            task_classes,
            task_links,
        )
        net = task_fact.make_network()

        # build arguments for the task list
        task_name_list = [
            "LandingLocationSelection",
            "LandingLocationPrep",
            "Fly",
            "LandingCheck",
            "Land",
            "MaintenanceWait",
        ]

        actor.add_task_network(net)
        # start the task network
        actor.start_network_loop("plane_net", init_task_name="GroundWait")
        env.run()
        assert env.now == 0

        new_actor = net.rehearse_network(actor=actor, task_name_list=task_name_list)

        # Did the new actor do what we wanted it to?
        base = new_actor.get_knowledge("base")
        base2 = actor.stage.world.bases[0]
        # Notice that due to copying the actor, the bases aren't exactly the same
        assert base.name == base2.name, "Wrong base selected"
        assert len(new_actor._knowledge) == 1, "Too much knowledge left"
        assert pytest.approx(new_actor.fuel, abs=0.01) == 86.199

        # Is the original actor untouched?
        assert len(actor._knowledge) == 0, "Actor should not have done anything"

        assert new_actor.code == 2, "Wrong MX code"

        assert new_actor.env.now > 2.0, "Cloned actor environment at the wrong time"

        task = actor.get_running_task("plane_net")
        assert task is not None and task.name == "GroundWait"


def test_rehearsing_from_actor() -> None:
    with EnvironmentContext() as env:
        actor = _build_test(env)
        task_fact = TaskNetworkFactory(
            "plane_net",
            task_classes,
            task_links,
        )
        net = task_fact.make_network()

        # build arguments for the task list
        task_name_list = [
            "LandingLocationSelection",
            "LandingLocationPrep",
            "Fly",
            "LandingCheck",
            "Land",
            "MaintenanceWait",
        ]

        actor.add_task_network(net)

        new_actor = actor.rehearse_network(
            "plane_net",
            task_name_list,
            knowledge={"dummy_know": 8675309},
        )

        # Make sure the knowledge was set on the copy, but not the original
        assert actor.get_knowledge("dummy_know") is None
        assert new_actor.get_knowledge("dummy_know") == 8675309

        # Did the new actor do what we wanted it to?
        base = new_actor.get_knowledge("base")
        base2 = actor.stage.world.bases[0]
        # Notice that due to copying the actor, the bases aren't exactly the same
        assert base.name == base2.name, "Wrong base selected"
        assert len(new_actor._knowledge) == 2, "Too much knowledge left"
        assert pytest.approx(new_actor.fuel, abs=0.01) == 86.199

        # Is the original actor untouched?
        assert len(actor._knowledge) == 0, "Actor should not have done anything"
        assert new_actor.code == 2, "Wrong MX code"

        assert new_actor.env.now > 2.0, "Cloned actor environment at the wrong time"


def test_running_simple_network() -> None:
    with EnvironmentContext() as env:
        actor = _build_test(env)
        task_fact = TaskNetworkFactory(
            "plane_net",
            task_classes,
            task_links,
        )
        net = task_fact.make_network()

        assert str(net) == "Task network: plane_net"

        # build arguments for the task list
        task_name_list = [
            "LandingLocationSelection",
            "LandingLocationPrep",
            "Fly",
            "LandingCheck",
            "Land",
            "MaintenanceWait",
        ]

        # tell the actor the queue its getting
        actor.add_task_network(net)
        actor.set_task_queue("plane_net", task_name_list)

        # run the queue with the network
        net.loop(actor=actor)
        env.run()

        base = actor.get_knowledge("base")
        base2 = actor.stage.world.bases[0]
        assert base is base2, "Wrong base selected"
        assert len(actor._knowledge) == 1, "Too much knowledge left"
        assert pytest.approx(actor.fuel, abs=0.01) == 86.199
        assert actor.code == 0, "Wrong MX code"


def test_interrupting_network() -> None:
    with EnvironmentContext() as env:
        actor = _build_test(env)
        task_fact = TaskNetworkFactory(
            "plane_net",
            task_classes,
            task_links,
        )
        net = task_fact.make_network()

        # build arguments for the task list
        task_name_list = [
            "LandingLocationSelection",
            "LandingLocationPrep",
            "Fly",
            "LandingCheck",
            "Land",
            "MaintenanceWait",
        ]

        # tell the actor the queue its getting
        actor.add_task_network(net)
        actor.set_task_queue("plane_net", task_name_list)

        # create a process that interrupts the plane during different times
        def interrupting_proc(
            env: Environment, actor: Aircraft, interrupt_time: float
        ) -> SIMPY_GEN:
            yield env.timeout(interrupt_time)
            # get the process
            network = actor._task_networks["plane_net"]
            assert network._current_task_proc is not None
            network._current_task_proc.interrupt(cause="a reason")

        # run the queue with the network
        net.loop(actor=actor)
        env.process(interrupting_proc(env, actor, 1.0))
        env.run()

        # the plane should land still
        assert actor._task_queue["plane_net"] == [], "Actor had tasks left"
        assert actor.code == 0, "Actor didn't get maintained"


def test_interrupting_network_with_cause() -> None:
    with EnvironmentContext() as env:
        actor = _build_test(env)
        task_fact = TaskNetworkFactory(
            "plane_net",
            task_classes,
            task_links,
        )
        net = task_fact.make_network()

        # build arguments for the task list
        task_name_list = [
            "LandingLocationSelection",
            "LandingLocationPrep",
            "Fly",
            "LandingCheck",
            "Land",
            "MaintenanceWait",
        ]

        # tell the actor the queue its getting
        actor.add_task_network(net)
        actor.set_task_queue("plane_net", task_name_list)

        # create a process that interrupts the plane during different times
        def interrupting_proc(
            env: Environment, actor: Aircraft, interrupt_time: float
        ) -> SIMPY_GEN:
            yield env.timeout(interrupt_time)
            # get the process
            # network = actor._task_networks["plane_net"]
            # network._current_task_proc.interrupt(cause="Code 4")
            actor.interrupt_network("plane_net", cause="Code 4")

        # run the queue with the network
        net.loop(actor=actor)
        env.process(interrupting_proc(env, actor, 1.0))
        env.run()


def test_interrupting_network_with_restart() -> None:
    with EnvironmentContext() as env:
        actor = _build_test(env)
        task_fact = TaskNetworkFactory(
            "plane_net",
            task_classes,
            task_links,
        )
        net = task_fact.make_network()

        # build arguments for the task list
        task_name_list = [
            "LandingLocationSelection",
            "LandingLocationPrep",
            "Fly",
            "LandingCheck",
            "Land",
            "MaintenanceWait",
        ]

        # tell the actor the queue its getting
        actor.add_task_network(net)
        actor.set_task_queue("plane_net", task_name_list)

        # create a process that interrupts the plane during different times
        def interrupting_proc(
            env: Environment, actor: Aircraft, interrupt_time: float
        ) -> SIMPY_GEN:
            yield env.timeout(interrupt_time)
            # get the process
            network = actor._task_networks["plane_net"]
            assert network._current_task_name is not None and network._current_task_name == "Fly"
            assert network._current_task_proc is not None
            network._current_task_proc.interrupt(cause="restart")

        # run the queue with the network
        net.loop(actor=actor)
        env.process(interrupting_proc(env, actor, 0.1))
        env.run()
        # the plane should land still
        assert actor._task_queue["plane_net"] == [], "Actor had tasks left"
        assert actor.code == 0, "Actor didn't get maintained"
        # it should take longer than the cancelled version
        assert pytest.approx(env.now, abs=0.0001) == 21.464767


def test_rehearsal_time() -> None:
    class Thing(Actor):
        the_time = LinearChangingState(recording=True)

    class ThingWait(Task):
        def task(self, *, actor: Thing) -> TASK_GEN:
            actor.activate_state(
                state="the_time",
                task=self,
                rate=1.0,
            )
            yield Wait.from_random_uniform(1.0, 2.0)
            actor.deactivate_all_states(task=self)

    with EnvironmentContext():
        tasks = {"ThingWait": ThingWait}
        task_links = {"ThingWait": TaskLinks(default="ThingWait", allowed=["ThingWait"])}
        factory = TaskNetworkFactory("fact", tasks, task_links)

        thing = Thing(name="Actor", the_time=0)
        thing.add_task_network(factory.make_network())
        new_thing = thing.rehearse_network("fact", ["ThingWait", "ThingWait"])
        assert new_thing.env.now == new_thing.the_time, "Bad rehearsal env time"


def test_decision_task_hold() -> None:
    # Test the conditions found in https://github.com/gtri/upstage/issues/35
    # Looks at zero time holds vs pass-through decision tasks

    # Test for new behavior first.
    data = []

    class Waiter(Task):
        def task(self, *, actor: Actor) -> TASK_GEN:
            data.append(f"{self.env.now:.1f} >> {actor.name} in Waiter")
            yield Wait(1.0)

    class Runner(Task):
        def task(self, *, actor: Actor) -> TASK_GEN:
            data.append(f"{self.env.now:.1f} >> {actor.name} in Runner")
            yield Wait(2.0)

    class Thinker(DecisionTask):
        DO_NOT_HOLD = True

        def make_decision(self, *, actor: Actor) -> None:
            data.append(f"{self.env.now:.1f} >> {actor.name} in Thinker")
            if "one" in actor.name:
                self.set_actor_task_queue(actor, ["Waiter"])
            else:
                self.set_actor_task_queue(actor, ["Runner"])

    net = TaskNetworkFactory(
        name="Example Net",
        task_classes={"Waiter": Waiter, "Runner": Runner, "Thinker": Thinker},
        task_links={
            "Waiter": TaskLinks(default="Thinker", allowed=["Thinker"]),
            "Thinker": TaskLinks(default="", allowed=["Waiter", "Runner"]),
            "Runner": TaskLinks(default="Thinker", allowed=["Thinker"]),
        },
    )
    with EnvironmentContext() as env:
        a = Actor(name="Actor one", debug_log=True)
        b = Actor(name="Actor two", debug_log=True)

        for actor in [a, b]:
            n = net.make_network()
            actor.add_task_network(n)
            actor.start_network_loop(n.name, "Waiter")

        env.run(until=2)

    expected = [
        "0.0 >> Actor one in Waiter",
        "0.0 >> Actor two in Waiter",
        "1.0 >> Actor one in Thinker",
        "1.0 >> Actor one in Waiter",
        "1.0 >> Actor two in Thinker",
        "1.0 >> Actor two in Runner",
    ]
    assert data == expected

    # Reset data in place, test for default behavior
    data[:] = []

    Thinker.DO_NOT_HOLD = False
    with EnvironmentContext() as env:
        a = Actor(name="Actor one", debug_log=True)
        b = Actor(name="Actor two", debug_log=True)

        for actor in [a, b]:
            n = net.make_network()
            actor.add_task_network(n)
            actor.start_network_loop(n.name, "Waiter")

        env.run(until=2)
    expected = [
        "0.0 >> Actor one in Waiter",
        "0.0 >> Actor two in Waiter",
        "1.0 >> Actor one in Thinker",
        "1.0 >> Actor two in Thinker",
        "1.0 >> Actor one in Waiter",
        "1.0 >> Actor two in Runner",
    ]
    assert data == expected



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""States that enable sharing between tasks."""

from upstage_des.actor import Actor
from upstage_des.base import UpstageError
from upstage_des.states import ActiveState
from upstage_des.task import Task


class SharedLinearChangingState(ActiveState[float]):
    """A state whose value changes linearly over time.

    Allows for multiple users of the state, keyed on actor and task.

    Assumes it's a non-frozen, floating-point value.

    Still activated in the usual way:

    >>> class Example(Actor):
    >>>     fuel = SharedLinearChangingState()
    ...
    >>> example.activate_state(
    >>>     state="fuel",
    >>>     task=self,
    >>>     rate=actor.fuel_burn,
    >>> )
    """

    def __init__(
        self,
        *,
        default: float | None = None,
        recording: bool = False,
    ) -> None:
        """Create a linear changing state that is shareable.

        Args:
            default (float | None, optional): Default value. Defaults to None.
            recording (bool, optional): If the state records. Defaults to False.
        """
        super().__init__(
            default=default,
            frozen=False,
            valid_types=float,
            recording=recording,
            default_factory=None,
        )
        self.IGNORE_LOCK: bool = True

    def _active(self, instance: Actor) -> float | None:
        """Return a value to set based on time or some other criteria.

        Args:
            instance (Actor): The actor instance of the state

        Returns:
            float: The value of the state
        """
        # Note the task needs to default to none, so when we call 'active'
        # to update the rate, we aren't adding a new one
        data = self.get_activity_data(instance)
        now: float = data["now"]
        current: float = data["value"]
        rate_tasks: dict[Task, float] = data.get("_rate_tasks", {})
        started_at: float | None = data.get("started_at", None)
        if started_at is None:
            # it's not currently active
            return None
        # no matter what, update the value
        curr_rate = sum(rate_tasks.values())

        last_calc_time: float = data.get("_last_time", now)
        elapsed = now - last_calc_time
        change = elapsed * curr_rate
        new_value = current + change

        # Task shows up from the activation
        if "task" in data:
            rate_to_add: float = data["rate"]
            task: Task = data["task"]
            if task in rate_tasks:
                raise UpstageError(
                    f"Duplicate task setting a rate {task}"
                    f"setting {self.name} on {instance}."
                    "You may have forgotten to deactivate."
                )
            rate_tasks[task] = rate_to_add

        self.__set__(instance, new_value)
        instance._set_active_state_data(
            state_name=self.name,
            started_at=now,
            _rate_tasks=rate_tasks,
            _last_time=now,
        )
        return new_value

    def deactivate(self, instance: Actor, task: Task | None = None) -> bool:
        """Deactivate the state.

        Args:
            instance (Actor): Actor the state is on
            task (Task): The task that is stopping its rate.

        Returns:
            bool: Still active or not
        """
        if task is None:
            raise UpstageError("Unexpected rate deactivation without a task.")
        data = self.get_activity_data(instance)
        rate_tasks: dict[Task, float] = data.get("_rate_tasks", {})
        # Since this state is ignoring the lock,
        # we can get here without it being active.
        if task not in rate_tasks:
            raise UpstageError(f"Task {task} is not changing {self.name} on {instance}")
        del rate_tasks[task]
        instance._set_active_state_data(
            state_name=self.name,
            _rate_tasks=rate_tasks,
        )
        # force a re-calculation
        getattr(instance, self.name)
        if rate_tasks:
            return True
        return False



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import upstage_des.api as UP
from upstage_des.type_help import TASK_GEN


class Act(UP.Actor):
    state = UP.State[int](default=0)


class Do(UP.Task):
    def task(self, *, actor: Act) -> TASK_GEN:
        yield UP.Wait(2)
        actor.state = actor.state + 1


class Do2(UP.Task):
    def task(self, *, actor: Act) -> TASK_GEN:
        yield UP.Wait(3)
        actor.state = actor.state + 3


def test_single_loop_and_names() -> None:
    with UP.EnvironmentContext() as env:
        factory = UP.TaskNetworkFactory.from_single_looping("example", Do)

        actor = Act(name="a thing")

        suggest = actor.suggest_network_name(factory)
        assert suggest == "example"
        new_net = factory.make_network(other_name=suggest)
        actor.add_task_network(new_net)

        suggest = actor.suggest_network_name(factory)
        assert suggest == "example_1"
        new_net = factory.make_network(other_name=suggest)
        actor.add_task_network(new_net)

        actor.start_network_loop("example", init_task_name="Do")
        actor.start_network_loop("example_1", init_task_name="Do")

        env.run(until=3)

        assert actor.state == 2

        assert actor.has_task_network("example")
        assert actor.has_task_network("example_1")

        actor.delete_task_network("example")
        actor.delete_task_network("example_1")

        assert not actor.has_task_network("example")
        assert not actor.has_task_network("example_1")

        env.run(until=5)

        assert actor.state == 4


def test_other_inits() -> None:
    with UP.EnvironmentContext() as env:
        factory = UP.TaskNetworkFactory.from_ordered_terminating("example", [Do, Do2])
        actor = Act(name="a thing")

        new_net = factory.make_network()
        actor.add_task_network(new_net)

        actor.start_network_loop("example", init_task_name="Do2")

        env.run(until=4)

        assert actor.state == 3

        task = actor.get_running_task("example")
        assert task is not None and "TERMINATING" in task.name

    with UP.EnvironmentContext() as env:
        factory = UP.TaskNetworkFactory.from_ordered_loop("example", [Do, Do2])
        actor = Act(name="a thing")

        new_net = factory.make_network()
        actor.add_task_network(new_net)

        actor.start_network_loop("example", init_task_name="Do")

        env.run(until=6)

        assert actor.state == 4



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""The task network class, and factory classes."""

from collections.abc import Generator, Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from upstage_des.actor import Actor

from simpy import Process

from upstage_des.base import SimulationError
from upstage_des.task import DecisionTask, Task, TerminalTask, process

REH_ACTOR = TypeVar("REH_ACTOR", bound="Actor")


@dataclass
class TaskLinks:
    """Type hinting for task link dictionaries."""

    default: str | None
    allowed: Sequence[str]


class TaskNetwork:
    """A means to represent, execute, and rehearse interdependent tasks."""

    def __init__(
        self,
        name: str,
        task_classes: Mapping[str, type[Task]],
        task_links: Mapping[str, TaskLinks],
    ) -> None:
        """Create a task network.

        Task links are defined as:
            {task_name: TaskLinks(default= task_name | None, allowed= list[task_names]}
        where each task has a default next task (or None), and tasks that could follow it.

        Args:
            name (str): Network name
            task_classes (Mapping[str, Task]): Task names to Task object mapping.
            task_links (Mapping[str, TaskLinks]): Task links.
        """
        self.name = name
        self.task_classes = task_classes
        self.task_links = task_links
        self._current_task_name: str | None = None
        self._current_task_inst: Task | None = None
        self._current_task_proc: Process | None = None

    def is_feasible(self, curr: str, new: str) -> bool:
        """Determine if a task can follow another one.

        Args:
            curr (str): Current task name
            new (str): Potential next task name

        Returns:
            bool: If the new task can follow the current.
        """
        value = self.task_links[curr].allowed
        return new in value

    def _next_task_name(
        self, curr_task_name: str, actor: "Actor", clear_queue: bool = False
    ) -> str:
        """Get the next task name.

        Returns:
            str: Task name
        """
        task_from_queue = actor.get_next_task(self.name)
        default_next_task = self.task_links[curr_task_name].default
        if task_from_queue is None:
            if default_next_task is None:
                raise SimulationError(  # pramga: no cover
                    f"No default task set for after {curr_task_name} on {actor}."
                )
            next_name = default_next_task
        else:
            next_name = task_from_queue
            # once we have the name, pop it from the queue
            if clear_queue:
                actor._clear_task(self.name)
        return next_name

    @process
    def loop(
        self, *, actor: "Actor", init_task_name: str | None = None
    ) -> Generator[Process, None, None]:
        """Start a task network running its loop.

        If no initial task name is given, it will default to following the queue.

        Args:
            actor (Actor): The actor to run the loop on.
            init_task_name (Optional[str], optional): Optional task to start running.
            Defaults to None.
        """
        next_name = actor.get_next_task(self.name)
        if next_name is None:
            if init_task_name is None:
                raise SimulationError(
                    f"Actor {actor} wasn't supplied an initial task"
                )  # pramga: no cover
            next_name = init_task_name

        self._current_task_name = next_name

        while True:
            task_name = self._current_task_name
            assert isinstance(task_name, str)
            actor.log(f"Outer: starting {task_name}")
            actor._begin_next_task(self.name, task_name)
            task_cls = self.task_classes[task_name]
            task_instance: Task = task_cls()
            self._current_task_inst = task_instance
            self._current_task_inst._set_network_name(self.name)
            self._current_task_inst._set_network_ref(self)

            if (
                isinstance(self._current_task_inst, DecisionTask)
                and self._current_task_inst.DO_NOT_HOLD
            ):
                self._current_task_inst.run_skip(actor=actor)
            else:
                self._current_task_proc = self._current_task_inst.run(actor=actor)
                yield self._current_task_proc

            next_name = self._next_task_name(task_name, actor)
            self._current_task_name = next_name

    def rehearse_network(
        self,
        *,
        actor: REH_ACTOR,
        task_name_list: list[str],
        knowledge: dict[str, Any] | None = None,
        end_task: str | None = None,
    ) -> REH_ACTOR:
        """Rehearse a path through the task network.

        Args:
            actor (Actor): The actor to perform the task rehearsal withs
            task_name_list (list[str]): The tasks to be performed in order
            knowledge (dict[str, Any], optional): Knowledge to give to the cloned/rehearsing actor
            end_task (str, optional): A task name to end on

        Returns:
            Actor: A copy of the original actor with state changes associated with the network.
        """
        _old_name = self._current_task_name
        _old_inst = self._current_task_inst
        _old_proc = self._current_task_proc
        knowledge = {} if knowledge is None else knowledge
        num_tasks = len(task_name_list)
        # pre-clone the actor to get a hold of the new environment
        new_actor = actor.clone(knowledge=knowledge)
        task_idx = 0
        while True:
            if task_idx < num_tasks:
                task_name = task_name_list[task_idx]
            elif end_task is None:
                break
            else:
                # Grab the default or one from the queue, clearing the queue to prevent loops
                task_name = self._next_task_name(task_name, new_actor, clear_queue=True)
            if end_task is not None and end_task == task_name:
                break  # pragma: no cover
            self._current_task_name = task_name
            self._current_task_inst = self.task_classes[task_name]()
            self._current_task_inst._set_network_name(self.name)
            new_actor = self._current_task_inst.rehearse(
                actor=new_actor,
                cloned_actor=True,
            )
            # The next name should be feasible
            if task_idx < num_tasks - 1:
                follow_on = task_name_list[task_idx + 1]
                if not self.is_feasible(task_name, follow_on):
                    raise SimulationError(  # pragma: no cover
                        f"Task {follow_on} not allowed after '{task_name}' in network"
                    )
            task_idx += 1
        # reset the internal parameters
        self._current_task_name = _old_name
        self._current_task_inst = _old_inst
        self._current_task_proc = _old_proc
        return new_actor

    def __repr__(self) -> str:
        return f"Task network: {self.name}"


class TaskNetworkFactory:
    """A factory for creating task network instances."""

    def __init__(
        self,
        name: str,
        task_classes: Mapping[str, type[Task]],
        task_links: Mapping[str, TaskLinks],
    ) -> None:
        """Create a factory for making instances of a task network.

        Task links are defined as:
            {task_name: TaskLinks(default= task_name | None, allowed= list[task_names]}
        where each task has a default next task (or None), and tasks that could follow it.

        Args:
            name (str): The network name
            task_classes (dict[str, Task]): Network task classes
            task_links (dict[str, dict[str, str  |  list[str]  |  None]]): Network links.
        """
        self.name = name
        self.task_classes = task_classes
        self.task_links = task_links

    @classmethod
    def from_single_looping(cls, name: str, task_class: type[Task]) -> "TaskNetworkFactory":
        """Create a network factory from a single task that loops.

        Args:
            name (str): Network name
            task_class (Task): The single task to loop

        Returns:
            TaskNetworkFactory: The factory for the single looping network.
        """
        taskname = task_class.__name__
        task_classes = {taskname: task_class}
        task_links: dict[str, TaskLinks] = {
            taskname: TaskLinks(default=taskname, allowed=[taskname])
        }
        return TaskNetworkFactory(name, task_classes, task_links)

    @classmethod
    def from_single_terminating(cls, name: str, task_class: type[Task]) -> "TaskNetworkFactory":
        """Create a network factory from a single task that terminates.

        Args:
            name (str): Network name
            task_class (Task): The single task to terminate after

        Returns:
            TaskNetworkFactory: The factory for the single terminating network.
        """
        taskname = task_class.__name__
        end_name = f"{taskname}_FINAL"
        task_classes = {taskname: task_class, end_name: TerminalTask}
        task_links: dict[str, TaskLinks] = {
            taskname: TaskLinks(default=end_name, allowed=[end_name])
        }
        return TaskNetworkFactory(name, task_classes, task_links)

    @classmethod
    def from_ordered_terminating(
        cls, name: str, task_classes: list[type[Task]]
    ) -> "TaskNetworkFactory":
        """Create a network factory from a list of tasks that terminates.

        Args:
            name (str): Network name
            task_classes (list[Task]): The tasks to run in order.

        Returns:
            TaskNetworkFactory: The factory for the ordered network.
        """
        task_class = {}
        task_links: dict[str, TaskLinks] = {}
        for i, tc in enumerate(task_classes):
            the_name = tc.__name__
            task_class[the_name] = tc
            try:
                nxt = task_classes[i + 1]
                nxt_name = nxt.__name__
            except IndexError:
                nxt = TerminalTask
                nxt_name = f"{name}_TERMINATING"
                task_class[nxt_name] = nxt
            task_links[the_name] = TaskLinks(default=nxt_name, allowed=[nxt_name])
        return TaskNetworkFactory(name, task_class, task_links)

    @classmethod
    def from_ordered_loop(cls, name: str, task_classes: list[type[Task]]) -> "TaskNetworkFactory":
        """Create a network factory from a list of tasks that loops.

        Args:
            name (str): Network name
            task_classes (list[Task]): The tasks to run in order.

        Returns:
            TaskNetworkFactory: The factory for the ordered network.
        """
        task_class = {}
        task_links: dict[str, TaskLinks] = {}
        for i, tc in enumerate(task_classes):
            the_name = tc.__name__
            task_class[the_name] = tc
            try:
                nxt = task_classes[i + 1]
            except IndexError:
                nxt = task_classes[0]
            nxt_name = nxt.__name__
            task_links[the_name] = TaskLinks(default=nxt_name, allowed=[nxt_name])
        return TaskNetworkFactory(name, task_class, task_links)

    def make_network(self, other_name: str | None = None) -> TaskNetwork:
        """Create an instance of the task network.

        By default, this uses the name defined on instantiation.

        Args:
            other_name (str, optional): Another name for the network. Defaults to None.

        Returns:
            TaskNetwork
        """
        use_name = other_name if other_name is not None else self.name
        return TaskNetwork(use_name, self.task_classes, self.task_links)



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Classes for UPSTAGE events that feed to simpy."""

from collections.abc import Callable
from typing import Any as tyAny
from warnings import warn

import simpy as SIM
from simpy.resources.container import ContainerGet, ContainerPut
from simpy.resources.resource import Release, Request
from simpy.resources.store import StoreGet, StorePut

from .base import SimulationError, UpstageBase, UpstageError
from .constants import PLANNING_FACTOR_OBJECT
from .units import unit_convert

__all__ = (
    "All",
    "Any",
    "BaseEvent",
    "Event",
    "Get",
    "FilterGet",
    "MultiEvent",
    "Put",
    "ResourceHold",
    "Wait",
)

SIM_REQ_EVTS = ContainerGet | ContainerPut | StoreGet | StorePut | Request | Release


class BaseEvent(UpstageBase):
    """Base class for framework events."""

    def __init__(self, *, rehearsal_time_to_complete: float = 0.0):
        """Create a base event with a notion of rehearsal time.

        Args:
            rehearsal_time_to_complete (float, optional): Time to simulate passing
                on rehearsal. Defaults to 0.0.
        """
        super().__init__()
        self._simpy_event: SIM.Event | None = None
        self._rehearsing: bool = False
        self._done_rehearsing: bool = False

        self.created_at: float = self.now
        self.rehearsal_time_to_complete = rehearsal_time_to_complete

    @property
    def now(self) -> float:
        """Current sim time.

        Returns:
            float: sim time
        """
        return self.env.now

    def calculate_time_to_complete(self) -> float:
        """Calculate the time elapsed until the event is triggered.

        Returns:
            float: The time until the event triggers.
        """
        return self.rehearsal_time_to_complete

    def as_event(self) -> SIM.Event:
        """Convert UPSTAGE event to a simpy Event.

        Returns:
            SIM.Event: The upstage event as a simpy event.
        """
        raise NotImplementedError(
            "Events must specify how to convert to :class:`simpy.events.Event`"
        )

    def is_complete(self) -> bool:
        """Is the event complete?

        Returns:
            bool: If it's complete or not.
        """
        if self._rehearsing:
            return self._done_rehearsing
        if self._simpy_event is None:
            raise UpstageError("Event has no simpy equivalent made.")
        return self._simpy_event.processed

    def cancel(self) -> None:
        """Cancel an event."""
        raise NotImplementedError("Implement custom event cancelling")

    @property
    def rehearsing(self) -> bool:
        """If the event is rehearsing.

        Returns:
            bool
        """
        return self._rehearsing

    @property
    def done_rehearsing(self) -> bool:
        """If the event is done rehearsing.

        Returns:
            bool
        """
        return self._done_rehearsing

    def _start_rehearsal(self) -> None:
        """Set the event to testing mode."""
        self._rehearsing = True
        self._done_rehearsing = False

    def _finish_rehearsal(self, complete: bool) -> None:
        """Finish rehearsing the event.

        Args:
            complete (bool): Indicates if the event was successful during the test.
        """
        if not self._rehearsing:
            raise SimulationError(
                "Trying to finish testing event but event testing was not started`"
            )
        self._done_rehearsing = complete

    def rehearse(self) -> tuple[float, tyAny | None]:
        """Run the event in 'rehearsal' mode without changing the real environment.

        This is used by the task rehearsal functions.

        Returns:
            tuple[float, Any | None]: The time to complete and the event's response.
        """
        self._start_rehearsal()
        time_advance = self.calculate_time_to_complete()
        self._finish_rehearsal(complete=True)

        event_response = None

        return time_advance, event_response


class Wait(BaseEvent):
    """Wait a specified or random uniformly distributed amount of time.

    Return a timeout. If time is a list of length 2, choose a random time
    between the interval given.

    Rehearsal time is given by the maximum time of the interval, if given.

    Parameters
    ----------
    timeout : int, float, list, tuple
        Amount of time to wait.  If it is a list or a tuple of length 2, a
        random uniform value between the two values will be used.

    """

    def _convert_time(self, time: float | int, unit: str | None) -> float:
        """Convert a time to the stage time.

        Args:
            time (float | int): The current time
            unit (str): Units the time is in

        Returns:
            float: Time in stage units
        """
        base_unit = self.stage.get("time_unit")
        if base_unit is not None and unit is not None:
            return unit_convert(time, unit, base_unit)
        return time

    def __init__(
        self,
        timeout: float | int,
        timeout_unit: str | None = None,
        *,
        rehearsal_time_to_complete: float | int | None = None,
    ) -> None:
        """Create a timeout event.

        If timeout_unit is specified, UPSTAGE will try to convert it to the
        time_unit set in the stage. Otherwise, it defaults to that time unit.

        Args:
            timeout (float | int): Time to wait.
            timeout_unit (str, optional): Units of time
            rehearsal_time_to_complete (float | int, optional): The rehearsal time
                to complete. Defaults to None (the timeout given).

        """
        if not isinstance(timeout, float | int):
            raise SimulationError("Bad timeout. Did you mean to use from_random_uniform?")
        timeout = self._convert_time(timeout, timeout_unit)
        self._time_to_complete = timeout
        self.timeout = timeout
        if self._time_to_complete < 0:
            raise SimulationError(f"Negative timeout in Wait: {self._time_to_complete}")
        rehearse = timeout if rehearsal_time_to_complete is None else rehearsal_time_to_complete
        super().__init__(rehearsal_time_to_complete=rehearse)
        self._simpy_event: SIM.Timeout | None = None

    @classmethod
    def from_random_uniform(
        cls,
        low: float | int,
        high: float | int,
        timeout_unit: str | None = None,
        *,
        rehearsal_time_to_complete: float | int | None = None,
    ) -> "Wait":
        """Create a wait from a random uniform time.

        If timeout_unit is specified, UPSTAGE will try to convert it to the
        time_unit set in the stage. Otherwise, it defaults to that time unit.

        Args:
            low (float): Lower bounds of random draw
            high (float): Upper bounds of random draw
            timeout_unit (str, optional): Units of time
            rehearsal_time_to_complete (float | int, optional): The rehearsal time
                to complete. Defaults to None - meaning the random value drawn.

        Returns:
            Wait: The timeout event
        """
        rng = UpstageBase().stage.random
        timeout = rng.uniform(low, high)
        return cls(timeout, timeout_unit, rehearsal_time_to_complete=rehearsal_time_to_complete)

    def as_event(self) -> SIM.Timeout:
        """Cast Wait event as a simpy Timeout event.

        Returns:
            SIM.Timeout
        """
        assert isinstance(self.env, SIM.Environment)
        if self._simpy_event is None:
            self._simpy_event = self.env.timeout(self._time_to_complete)
        return self._simpy_event

    def cancel(self) -> None:
        """Cancel the timeout.

        There's no real meaning to cancelling a timeout. It sits in simpy's queue either way.
        """
        assert self._simpy_event is not None
        try:
            self._simpy_event.defused = True
        except RuntimeError as exc:
            warn(f"Runtime error when cancelling '{self}', Error: {exc}!")


class BaseRequestEvent(BaseEvent):
    """Base class for Request Events.

    Requests are things like Get and Put that wait in a queue.
    """

    def __init__(self, rehearsal_time_to_complete: float = 0.0) -> None:
        """Create a request event.

        Args:
            rehearsal_time_to_complete (float, optional): Estimated time to complete.
                Defaults to 0.0.
        """
        super().__init__(rehearsal_time_to_complete=rehearsal_time_to_complete)
        self._request_event: SIM_REQ_EVTS | None = None

    def cancel(self) -> None:
        """Cancel the Request."""
        if self._request_event is None:
            return
        if not self.is_complete():
            self._request_event.cancel()
        # TODO: Do we put it back?

    def is_complete(self) -> bool:
        """Test if the request is finished.

        Returns:
            bool
        """
        if self.rehearsing:
            if self.done_rehearsing is None:
                raise SimulationError(
                    f"Event '{self}' rehearsal started, but completion was"
                    "not set as incomplete, i.e., to `False`!"
                )
            return self.done_rehearsing
        assert self._request_event is not None
        return self._request_event.processed


class Put(BaseRequestEvent):
    """Wrap the ``simpy`` Put event.

    This is an event that puts an object into a ``simpy`` store or puts
    an amount into a container.

    """

    def __init__(
        self,
        put_location: SIM.Container | SIM.Store,
        put_object: float | int | tyAny,
        rehearsal_time_to_complete: float = 0.0,
    ) -> None:
        """Create a Put request for a store or container.

        Args:
            put_location (SIM.Container | SIM.Store): Any container, store, or subclass.
            put_object (float | int | Any): The amount (float | int) or object (Any) to put.
            rehearsal_time_to_complete (float, optional): Estimated time for the put to finish.
            Defaults to 0.0.
        """
        super().__init__(rehearsal_time_to_complete=rehearsal_time_to_complete)

        if not issubclass(put_location.__class__, SIM.Container | SIM.Store):
            raise SimulationError(
                f"put_location must be a subclass of Container "
                f"or Store, not {put_location.__class__}"
            )

        self.put_location = put_location
        self.put_object = put_object
        self._request_event: ContainerPut | StorePut | None = None

    def as_event(self) -> ContainerPut | StorePut:
        """Convert event to a ``simpy`` Event.

        Returns:
        ---------
        :obj:`simpy.events.Event`
            Put request as a simpy event.

        """
        if self._request_event is None:
            self._request_event = self.put_location.put(self.put_object)
        return self._request_event


class MultiEvent(BaseEvent):
    """A base class for evaluating multiple events.

    Note:
        Subclasses of MultiEvent must define these methods:
            * aggregation_function: Callable[[list[float]], float]
            * simpy_equivalent: simpy.Event

        For an example, refer to :class:`~Any` and :class:`~All`.
    """

    def __init__(self, *events: BaseEvent | SIM.Process) -> None:
        """Create a multi-event based on a list of events.

        Args:
            *events (BaseEvent): The events that comprise the multi-event.
        """
        super().__init__()

        for event in events:
            if not issubclass(event.__class__, BaseEvent):
                warn(
                    f"Event '{event}' is not an upstage Event. "
                    f"All events in a MultiEvent must be an "
                    f"instance of upstage BaseEvent if you are going "
                    f"to rehearse the task that contains this MultiEvent.",
                    UserWarning,
                )
        self.events = events
        self._simpy_event = None

    @staticmethod
    def aggregation_function(times: list[float]) -> float:
        """Aggregate event times to one single time.

        Args:
            times (list[float]): Event rehearsal times

        Returns:
            float: The aggregated time
        """
        raise NotImplementedError("Implement in subclass")

    @staticmethod
    def simpy_equivalent(env: SIM.Environment, events: list[SIM.Event]) -> SIM.Event:
        """Return the simpy equivalent event.

        Args:
            env (SIM.Environment): The SimPy environment.
            events (list[BaseEvent]): Events to turn into multi-event.

        Returns:
            SIM.Event: The aggregate event.
        """
        raise NotImplementedError("Implement in subclass")

    def _make_event(self, event: BaseEvent | SIM.Process) -> SIM.Event:
        # handle a process in the MultiEvent for non-rehearsal uses
        if isinstance(event, SIM.Process):
            return event
        return event.as_event()

    def as_event(self) -> SIM.Event:
        """Convert the UPSTAGE event to simpy.

        Returns:
            SIM.Event: typically an Any or All
        """
        sub_events = [self._make_event(event) for event in self.events]
        assert isinstance(self.env, SIM.Environment)
        self._simpy_event = self.simpy_equivalent(self.env, sub_events)
        return self._simpy_event

    def cancel(self) -> None:
        """Cancel the multi event and propagate it to the sub-events."""
        if self._simpy_event is None:
            raise UpstageError("Can't cancel a nonexistent event.")
        self._simpy_event.defused = True
        self._simpy_event.fail(Exception("defused"))
        for event in self.events:
            if isinstance(event, BaseEvent):
                event.cancel()

    def calculate_time_to_complete(
        self,
    ) -> float:
        """Compute time required to complete the multi-event.

        Args:
            return_sub_events (bool, Optional): Whether to return all times or not.
                Defaults to False.
        """
        event_times = {
            event: event.calculate_time_to_complete()
            for event in self.events
            if isinstance(event, BaseEvent)
        }

        time_to_complete = self.aggregation_function(list(event_times.values()))

        return time_to_complete

    def calc_time_to_complete_with_sub(self) -> tuple[float, dict[BaseEvent, float]]:
        """Compute time required for MultiEvent and get sub-event times.

        Returns:
            tuple[float, dict[BaseEvent, float]]: Aggregate and individual times.
        """
        event_times = {
            event: event.calculate_time_to_complete()
            for event in self.events
            if isinstance(event, BaseEvent)
        }
        time_to_complete = self.aggregation_function(list(event_times.values()))

        return time_to_complete, event_times

    def _start_rehearsal(self) -> None:
        """Start rehearsing all the sub-events."""
        super()._start_rehearsal()
        for event in self.events:
            if not hasattr(event, "_start_rehearsal"):
                raise SimulationError(
                    f"Event '{event}' is not an upstage Event. "
                    f"All events in a MultiEvent must be an "
                    f"instance of upstage BaseEvent if you are going"
                    f"to rehearse the task that contains this MultiEvent."
                )
            event._start_rehearsal()

    def rehearse(self) -> tuple[float, tyAny]:
        """Run the event in 'trial' mode without changing the real environment.

        Returns:
            tuple[float, Any]: The time to complete and the event's response.

        Note:
            This is used by the task rehearsal functions.
        """
        self._start_rehearsal()

        event_response = None
        time_to_finish, event_times = self.calc_time_to_complete_with_sub()

        for event, event_end_time in event_times.items():
            event._finish_rehearsal(complete=event_end_time <= time_to_finish)

        self._finish_rehearsal(complete=True)

        return time_to_finish, event_response


class Any(MultiEvent):
    """An event that requires one event to succeed before succeeding."""

    @staticmethod
    def aggregation_function(times: list[float]) -> float:
        """Aggregation function for rehearsal time.

        Args:
            times (list[float]): List of rehearsal times

        Returns:
            float: Aggregated time (the minimum)
        """
        return min(times)

    @staticmethod
    def simpy_equivalent(env: SIM.Environment, events: list[SIM.Event]) -> SIM.Event:
        """Return the SimPy version of the UPSTAGE Any event.

        Args:
            env (SIM.Environment): SimPy Environment.
            events (list[SIM.Event]): List of events.

        Returns:
            SIM.Event: A simpy AnyOf event.
        """
        return SIM.AnyOf(env, events)


class Get(BaseRequestEvent):
    """Wrap the ``simpy`` Get event.

    Event that gets an object from a ``simpy`` store or gets an amount from a
    container.
    """

    def __init__(
        self,
        get_location: SIM.Store | SIM.Container,
        *get_args: tyAny,
        rehearsal_time_to_complete: float = 0.0,
        **get_kwargs: tyAny,
    ) -> None:
        """Create a Get request on a store, container, or subclass of those.

        Args:
            get_location (SIM.Store | SIM.Container): The place for the Get request
            rehearsal_time_to_complete (float, optional): _description_. Defaults to 0.0.
            get_args (Any): optional positional args for the get request
                (blank for Store and Container)
            get_kwargs (Any): optional keyword args for the get request
                (blank for Store and Container)
        """
        super().__init__(rehearsal_time_to_complete=rehearsal_time_to_complete)

        if not issubclass(get_location.__class__, SIM.Container | SIM.Store):
            raise SimulationError(
                "'put_location' must be a subclass of Container"
                f" or Store, not {get_location.__class__}"
            )

        self.get_location = get_location
        self.get_args = get_args
        self.get_kwargs = get_kwargs
        self.__is_store = issubclass(get_location.__class__, SIM.Store)
        self._request_event: ContainerGet | StoreGet | None = None

    def calculate_time_to_complete(self) -> float:
        """Calculate time elapsed until the event is triggered.

        Returns:
            float: Estimated time until the event triggers.

        """
        return self.rehearsal_time_to_complete

    def as_event(self) -> ContainerGet | StoreGet:
        """Convert get to a ``simpy`` Event.

        Returns:
            ContainerGet | StoreGet
        """
        # TODO: optional checking for container types for feasibility
        if self._request_event is None:
            self._request_event = self.get_location.get(
                *self.get_args,
                **self.get_kwargs,
            )
        return self._request_event

    def get_value(self) -> tyAny:
        """Get the value returned when the request is complete.

        Returns:
            Any: The amount or item requested.
        """
        if self.__is_store:
            if self.rehearsing and self.done_rehearsing:
                return PLANNING_FACTOR_OBJECT
            if self._request_event is not None and self._request_event.value is not None:
                return self._request_event.value
            else:
                raise SimulationError("Requested item from an unfinished Get request.")
        else:
            raise SimulationError(
                "'get_value' is not supported for Containers. Check is_"
                "complete and use the amount you requested."
            )

    def rehearse(self) -> tuple[float, tyAny]:
        """Mock the event to test if it is feasible.

        Note:
            The function does not fully test the conditions to satisfy the
            get request, but this method can be called as part of a more
            complex rehearse run.

        Returns:
            float: The time it took to do the request
            Any: The value of the request.
        """
        time_advance, _ = super().rehearse()
        event_response = None
        if self.__is_store:
            event_response = PLANNING_FACTOR_OBJECT
        return time_advance, event_response


class ResourceHold(BaseRequestEvent):
    """Wrap the ``simpy`` request resource event.

    This manages getting and giving back all in one object.

    Example:
        >>> resource = simpy.Resource(env, capacity=1)
        >>> hold = ResourceHold(resource)
        >>> # yield on the hold to get it
        >>> yield hold
        >>> # now that you have it, do things..
        >>> # give it back
        >>> yield hold
        >>> ...
    """

    def __init__(
        self,
        resource: SIM.Resource,
        *resource_args: tyAny,
        rehearsal_time_to_complete: float = 0.0,
        **resource_kwargs: tyAny,
    ) -> None:
        """Create an event to use twice to get and give back a resource.

        Args:
            resource (SIM.Resource): The simpy resource object.
            rehearsal_time_to_complete (float, optional): Expected time to wait to
                get the resource. Defaults to 0.0.
            *resource_args (Any): positional arguments to the resource
            **resource_kwargs (Any): keyword arguments to the resource
        """
        super().__init__(rehearsal_time_to_complete=rehearsal_time_to_complete)

        self.resource = resource
        self.resource_args = resource_args
        self.resource_kwargs = resource_kwargs
        self._stage = "request"
        self._request: Request | Release | None = None

    def calculate_time_to_complete(self) -> float:
        """Time to complete, based on waiting for getting or giving back.

        Returns:
            float: Time
        """
        if self._stage == "request":
            # assume the stage will switch on the next call
            self._stage = "release"
            return self.rehearsal_time_to_complete
        elif self._stage == "release":
            return 0.0
        raise UpstageError(f"Resource request stage is wrong: {self._stage}")

    def as_event(self) -> Request | Release:
        """Create the simpy event for the right state of Resource usage.

        Returns:
            Request | Release: The simpy event.
        """
        if self._stage == "request":
            self._request = self.resource.request(*self.resource_args, **self.resource_kwargs)

            self._request_event = self._request
            self._stage = "release"
            return self._request_event
        elif self._stage == "release":
            if not self._request or not self._request.processed:
                raise SimulationError(
                    "Resource release requested when the "
                    "resource hasn't been given. Did you cancel?"
                )
            assert isinstance(self._request, Request)
            self._request_event = self.resource.release(self._request)
            self._stage = "completed"
            return self._request_event
        raise UpstageError(f"Bad stage for Resource Hold: {self._stage}")


class FilterGet(Get):
    """A Get for a FilterStore."""

    def __init__(
        self,
        get_location: SIM.FilterStore,
        filter: Callable[[tyAny], bool],
        rehearsal_time_to_complete: float = 0.0,
    ) -> None:
        """Create a Get request on a FilterStore.

        The filter function returns a boolean (in/out of consideration).

        Args:
            get_location (SIM.Store | SIM.Container): The place for the Get request
            filter (Callable[[Any], bool]): The function that filters items in the store
            rehearsal_time_to_complete (float, optional): _description_. Defaults to 0.0.
        """
        super().__init__(
            get_location=get_location,
            rehearsal_time_to_complete=rehearsal_time_to_complete,
            filter=filter,
        )


class All(MultiEvent):
    """An event that requires all events to succeed before succeeding."""

    @staticmethod
    def aggregation_function(times: list[float]) -> float:
        """Aggregate event times for rehearsal.

        Args:
            times (list[float]): List of rehearsing times.

        Returns:
            float: Aggregated (maximum) time.
        """
        return max(times)

    @staticmethod
    def simpy_equivalent(env: SIM.Environment, events: list[SIM.Event]) -> SIM.Event:
        """Return the SimPy version of the UPSTAGE All event.

        Args:
            env (SIM.Environment): SimPy Environment.
            events (list[SIM.Event]): List of events.

        Returns:
            SIM.Event: A simpy AllOf event.
        """
        return SIM.AllOf(env, events)


class Event(BaseEvent):
    """An UPSTAGE version of the standard SimPy Event.

    Returns a planning factor object on rehearsal for user testing against in rehearsals, in case.

    When the event is succeeded, a payload can be added through kwargs.

    This Event assumes that it might be long-lasting, and will auto-reset when yielded on.
    """

    def __init__(
        self,
        rehearsal_time_to_complete: float = 0.0,
        auto_reset: bool = True,
    ) -> None:
        """Create an event.

        Args:
            rehearsal_time_to_complete (float, optional): Expected time to complete.
                Defaults to 0.0.
            auto_reset (bool, optional): Whether to auto-reset on yield. Defaults to True.
        """
        super().__init__(rehearsal_time_to_complete=rehearsal_time_to_complete)
        # The usage is sometimes that events might succeed before being
        # yielded on
        self._payload: dict[str, Any] = {}
        self._auto_reset = auto_reset
        assert isinstance(self.env, SIM.Environment)
        self._event = SIM.Event(self.env)

    def calculate_time_to_complete(self) -> float:
        """Return the time to complete.

        Returns:
            float: Time to complete estimate.
        """
        return self.rehearsal_time_to_complete

    def as_event(self) -> SIM.Event:
        """Get the Event as a simpy type.

        This resets the event if allowed.

        Returns:
            SIM.Event
        """
        if self.is_complete():
            if self._auto_reset:
                self.reset()
            else:
                raise UpstageError("Event not allowed to reset on yield.")
        return self._event

    def succeed(self, **kwargs: tyAny) -> None:
        """Succeed the event and store any payload.

        Args:
            **kwargs (Any): key:values to store as payload.
        """
        if self.is_complete():
            raise SimulationError("Event has already completed")
        self._payload = kwargs
        self._event.succeed()

    def is_complete(self) -> bool:
        """Is the event done?

        Returns:
            bool
        """
        return self._event.processed

    def get_payload(self) -> dict[str, tyAny]:
        """Get any payload from the call to succeed().

        Returns:
            dict[str, Any]: The payload left by the succeed() caller.
        """
        return self._payload

    def reset(self) -> None:
        """Reset the event to allow it to be held again."""
        assert isinstance(self.env, SIM.Environment)
        self._event = SIM.Event(self.env)

    def cancel(self) -> None:
        """Cancel the event.

        Cancelling doesn't mean much, since it's still going to be yielded on.
        """
        try:
            self._event.defused = True
            self._event.succeed()
        except RuntimeError as exc:
            exc.add_note(f"Runtime error when cancelling '{self}'")
            raise exc

    def rehearse(self) -> tuple[float, tyAny]:
        """Run the event in 'trial' mode without changing the real environment.

        Returns:
            tuple[float, Any]: The time to complete and the event's response.
        """
        time_advance, _ = super().rehearse()
        return time_advance, PLANNING_FACTOR_OBJECT



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Communications process helper."""

from collections.abc import Callable, Generator
from typing import Any

from simpy import Event, Process, Store

from upstage_des.communications.comms import CommsManager, Message, MessageContent
from upstage_des.task import process


def generate_comms_wait(
    incoming_store: Store,
    callback: Callable[[MessageContent], Any],
) -> Callable[[], Process]:
    """Create a process function to transfer communications to a callback.

    This hides cleanup and other stability functions from the user.

    Parameters
    ----------
    incoming_store : A simpy or upstage store
        The store that is linked to a CommsManager instance.
    callback : function
        The function to call with a received message

    Returns:
    -------
    function
        An UPSTAGE process function that passes messages
    """

    @process
    def comms_wait_proc() -> Generator[Event, str | Message, None]:
        while True:
            message = yield incoming_store.get()
            message = CommsManager.clean_message(message)
            callback(message)

    return comms_wait_proc



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""The file contains the Nucleus features of UPSTAGE."""

from collections import defaultdict
from typing import Any

from upstage_des.actor import Actor
from upstage_des.base import UpstageError
from upstage_des.task_network import TaskNetwork


class NucleusInterrupt:
    """A data container for interrupting nucleus events."""

    def __init__(self, name: str, value: Any) -> None:
        """A container for Nucleus interrupt data.

        Args:
            name (str): Name of the state causing the interrupt.
            value (Any): The state value
        """
        self.state_name = name
        self.value = value

    def __repr__(self) -> str:
        return f"NucleusInterrupt: {self.state_name} {self.value}"


class TaskNetworkNucleus:
    """The nucleus, for state-based task network signaling."""

    def _attach(self) -> None:
        """Attach the nucleus to an actor."""
        self._actor.log(f"Attaching {self} as a state listener!")
        if self._actor._state_listener is not None:
            raise UpstageError(f"{self._actor} already has a nucleus attached.")
        self._actor._state_listener = self

    def __init__(
        self,
        *,
        actor: Actor,
    ) -> None:
        """Create a task network nucleus on an Actor.

        Args:
            actor (Actor): The actor instance.
        """
        self._actor = actor
        self._state_map: dict[str, set[str]] = defaultdict(set)
        self._network_map: dict[str, set[str]] = defaultdict(set)
        self._attach()

    def add_network(
        self,
        network_name: str | TaskNetwork,
        watch_states: list[str],
    ) -> None:
        """Add a network to the nucleus for state management.

        Args:
            network_name (str | TaskNetwork):  A task network that works on this nucleus/actor
            watch_states (list[str]): States that - when changed - cause the network to change.
        """
        if isinstance(network_name, TaskNetwork):
            network_name = network_name.name
        if network_name not in self._actor._task_networks:
            raise UpstageError(f"No network {network_name} in {self._actor}")
        for state in watch_states:
            self._state_map[state].add(network_name)
            self._network_map[network_name].add(state)
            # if not hasattr(actor, state):
            #     raise SimulationError(f"State {state} does not exist on actor")

    def remove_network(
        self,
        network_name: str | TaskNetwork,
    ) -> None:
        """Remove a network from nucleus.

        Args:
            network_name (str | TaskNetwork): A task network that works on this nucleus/actor
        """
        if isinstance(network_name, TaskNetwork):
            network_name = network_name.name
        if network_name not in self._actor._task_networks:
            raise UpstageError(f"No network {network_name} in {self._actor}")
        for state in self._network_map[network_name]:
            self._state_map[state].remove(network_name)
        del self._network_map[network_name]

    def send_change(self, state_name: str, state_value: Any) -> None:
        """Send a change notification for a given state.

        Args:
            state_name (str): The state's name.
            state_value (Any): The value of the state.
        """
        for net_name in self._state_map.get(state_name, []):
            net: TaskNetwork | None = self._actor._task_networks.get(net_name, None)
            if net is None:
                raise UpstageError(f"No network {net_name} in {self._actor}")
            proc = net._current_task_proc
            if proc is not None:
                proc.interrupt(
                    cause=NucleusInterrupt(state_name, state_value),
                )



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Geodetic frame conversions."""

from math import atan2, cos, degrees, radians, sin, sqrt

from .geo_types import POSITION, POSITIONS

# Constants for the earth
spherical_radius: float = 6378137.0
WGS84_A: float = 6378137.0  # meters - semimajor axis
WGS84_F: float = 1 / 298.257223563  # flattening factor
WGS84_B: float = WGS84_A * (1 - WGS84_F)  # meters - semiminor axis

# build the constant values
sR_e = spherical_radius
sR_p = spherical_radius
se_2 = (sR_e**2 - sR_p**2) / sR_e**2
sep_2 = (sR_e**2 - sR_p**2) / sR_p**2
se = sqrt(se_2)
sep = sqrt(se)

wR_e = WGS84_A
wR_p = WGS84_B
# first eccentricity squared
we_2 = WGS84_F * (2 - WGS84_F)
# second eccentricity squared
wep_2 = we_2 / (1 - WGS84_F) ** 2
we = sqrt(we_2)
wep = sqrt(we)
wF = WGS84_F


class BaseConversions:
    """Base class for converting in geodetic frames."""

    radius_e = 0.0
    radius_p = 0.0
    e_2 = 0.0
    ep_2 = 0.0
    e = 0.0
    ep = 0.0
    f = 0.0

    @classmethod
    def _lla2ecef(cls, lla: POSITION, input_in_radians: bool = False) -> POSITION:
        """Convert Lat-Lon-Altitude into ECEF.

        Args:
            lla (POSITION): lat/lon/alt data in 3 columns and 1 row.
            input_in_radians (bool, optional): If the lat/lon are in radians. Defaults to False.

        Returns:
            POSITION: XYZ ECEF value for the position.
        """
        rad_lat, rad_lon, ht = lla
        if not input_in_radians:
            rad_lat = radians(rad_lat)  # sometimes denoted as lambda
            rad_lon = radians(rad_lon)  # sometimes denoted as phi

        N = cls.radius_e / sqrt(1 - (cls.e_2 * sin(rad_lat) ** 2))
        temp1 = (N + ht) * cos(rad_lat)
        temp2 = N * (1 - cls.f) ** 2 + ht
        x = temp1 * cos(rad_lon)
        y = temp1 * sin(rad_lon)
        z = temp2 * sin(rad_lat)
        return (x, y, z)

    @classmethod
    def lla2ecef(cls, lla: POSITIONS, input_in_radians: bool = False) -> POSITIONS:
        """Convert Lat-Lon-Altitude into ECEF.

        Args:
            lla (POSITIONS): lat/lon/alt data in 3 columns and N rows.
            input_in_radians (bool, optional): If the lat/lon are in radians. Defaults to False.

        Returns:
            POSITIONS: Array of XYZ ECEF values.
        """
        return [cls._lla2ecef(row, input_in_radians) for row in lla]

    @classmethod
    def _ecef2lla(cls, ecef: POSITION, radians_out: bool = False) -> POSITION:
        """Convert ECEF to Lat/Lon/Altitude.

        Args:
            ecef (POSITION): ECEF point
            radians_out (bool, optional): If lat/lon out should be in radians. Defaults to False.

        Returns:
            POSITION: Lat/Lon/Altitude
        """
        X, Y, Z = ecef
        # Longitude (rad; 1xN)
        lon = atan2(Y, X)

        # Calculations
        e2 = cls.e_2
        r2 = X**2 + Y**2
        r = sqrt(r2)
        a = cls.radius_e
        b = cls.radius_p
        a2 = a**2
        b2 = b**2
        E2 = a2 - b2
        Z2 = Z**2
        F = 54 * b2 * Z2
        G = r2 + (1 - e2) * Z2 - e2 * E2
        c = (e2 * e2 * F * r2) / (G * G * G)

        s = (1 + c + sqrt(c * c + 2 * c)) ** (1 / 3)
        P = F / (3 * (s + 1 / s + 1) ** 2 * G * G)
        Q = sqrt(1 + 2 * e2 * e2 * P)
        ro = -(e2 * P * r) / (1 + Q) + sqrt(
            (a * a / 2) * (1 + 1 / Q) - ((1 - e2) * P * Z2) / (Q * (1 + Q)) - P * r2 / 2
        )
        tmp = (r - e2 * ro) ** 2
        U = sqrt(tmp + Z2)
        V = sqrt(tmp + (1 - e2) * Z2)
        zo = (b2 * Z) / (a * V)

        h = U * (1 - b2 / (a * V))
        lat = atan2((Z + cls.ep_2 * zo), r)

        if not radians_out:
            lon = degrees(lon)
            lat = degrees(lat)

        lla = (lat, lon, h)
        return lla

    @classmethod
    def ecef2lla(cls, ecef: POSITIONS, radians_out: bool = False) -> POSITIONS:
        """Convert ECEF to Lat/Lon/Altitude.

        Args:
            ecef (POSITIONS): ECEF points
            radians_out (bool, optional): If lat/lon out should be in radians. Defaults to False.

        Returns:
            POSITIONS: Lat/Lon/Altitude
        """
        return [cls._ecef2lla(row, radians_out) for row in ecef]


class SphericalConversions(BaseConversions):
    """Conversions on a spherical globe."""

    radius_e = sR_e
    radius_p = sR_p
    e_2 = se_2
    ep_2 = sep_2
    e = se
    ep = sep

    @classmethod
    def _lla2ecef(cls, lla: POSITION, input_in_radians: bool = False) -> POSITION:
        """Lat-Lon-Alt to ECEF.

        Args:
            lla (POSITION): Lat-Lon-Altitude points
            input_in_radians (bool, optional): If lat/lon are in radians. Defaults to False.

        Returns:
            POSITION: ECEF
        """
        rad_lat = lla[0]
        rad_lon = lla[1]
        if not input_in_radians:
            rad_lat = radians(rad_lat)  # sometimes denoted as lambda
            rad_lon = radians(rad_lon)  # sometimes denoted as phi

        ht = lla[2]
        rad = cls.radius_e + ht
        clat = cos(rad_lat)
        x = rad * cos(rad_lon) * clat
        y = rad * clat * sin(rad_lon)
        z = rad * sin(rad_lat)
        return (x, y, z)

    @classmethod
    def lla2ecef(cls, lla: POSITIONS, input_in_radians: bool = False) -> POSITIONS:
        """Lat-Lon-Alt to ECEF.

        Args:
            lla (POSITIONS): Lat-Lon-Altitude points
            input_in_radians (bool, optional): If lat/lon are in radians. Defaults to False.

        Returns:
            POSITIONS: ECEF
        """
        return [cls._lla2ecef(row, input_in_radians) for row in lla]

    @classmethod
    def _ecef2lla(cls, ecef: POSITION, radians_out: bool = False) -> POSITION:
        """Convert ECEF to Lat-Lon-Alt.

        Args:
            ecef (POSITION): Points in ECEF
            radians_out (bool, optional): If Lat/Lon out are in radians. Defaults to False.

        Returns:
            POSITION: Lat-Lon-Alt points
        """
        x, y, z = ecef
        # Longitude (rad)
        lon = atan2(y, x)

        p = sqrt(x**2 + y**2)
        # Latitude (rads)
        lat = atan2(z, p)
        h = p / cos(lat) - cls.radius_e

        # correct for numerical instability in altitude near exact poles:
        # (after this correction, error is about 2 millimeters, which is about
        # the same as the numerical precision of the overall function)
        k = (abs(x) < 1) & (abs(y) < 1)
        if k:
            h = abs(z) - cls.radius_p

        if not radians_out:
            lon = degrees(lon)
            lat = degrees(lat)

        lla = (lat, lon, h)
        return lla

    @classmethod
    def ecef2lla(cls, ecef: POSITIONS, radians_out: bool = False) -> POSITIONS:
        """Convert ECEF to Lat-Lon-Alt.

        Args:
            ecef (POSITIONS): Points in ECEF
            radians_out (bool, optional): If Lat/Lon out are in radians. Defaults to False.

        Returns:
            POSITIONS: Lat-Lon-Alt points
        """
        return [cls._ecef2lla(row, radians_out) for row in ecef]


class WGS84Conversions(BaseConversions):
    """WGS84 coordinate conversions."""

    radius_e = wR_e
    radius_p = wR_p
    e_2 = we_2
    ep_2 = wep_2
    e = we
    ep = wep
    f = wF



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""This file contains a queueing motion manager for sensor/mover intersections."""

from collections.abc import Callable, Generator
from typing import Any, Protocol, TypeVar
from warnings import warn

from simpy import Event as SimpyEvent
from simpy import Interrupt, Process

from upstage_des.actor import Actor
from upstage_des.base import (
    MotionAndDetectionError,
    SimulationError,
    UpstageBase,
)
from upstage_des.data_types import CartesianLocation, GeodeticLocation
from upstage_des.states import CartesianLocationChangingState, GeodeticLocationChangingState

VALID = [
    ("ENTER", "END_INSIDE"),
    ("ENTER", "EXIT"),
    ("START_INSIDE", "END_INSIDE"),
    ("START_INSIDE", "EXIT"),
]

LOC_TYPES = CartesianLocation | GeodeticLocation
LOC_LIST = list[CartesianLocation] | list[GeodeticLocation]

LOC_INPUT = TypeVar("LOC_INPUT", "GeodeticLocation", "CartesianLocation")

INTERSECTION_TIMING_CALLABLE = Callable[
    [
        LOC_INPUT,
        LOC_INPUT,
        float,
        LOC_INPUT,
        float,
    ],
    tuple[
        list[LOC_INPUT],
        list[float],
        list[str],
        float,
    ],
]


class SensorType(Protocol):
    """Protocol class for sensor typing."""

    def entity_exited_range(
        self,
        entity: Any,
    ) -> None:
        """Entity exit range and does something."""

    def entity_entered_range(
        self,
        entity: Any,
    ) -> None:
        """Entity enters range and does something."""


class SensorMotionManager(UpstageBase):
    """Schedules the interaction of moving and detectable entities against non-moving 'sensors'.

    Movable objects must be Actors with:
    1. GeodeticLocationChangingState OR CartesianLocationChangingState
    2. DetectabilityState

    Sensor objects MUST implement these two methods:
    1. `entity_entered_range(mover)`
    2. `entity_exited_range(mover)`

    The first is called when a mover enters the sensor's visiblity.
    The second is called when a mover leaves the visibility or becomes undetectable.

    The motion manager will learn about sensor objects with:

    sensor_motion_manager.add_sensor(sensor_object, location, radius)

    Where location is a location object found in upstage.data_types and radius
    is a distance in the units defined in upstage.STAGE.

    """

    def __init__(
        self, intersection_model: INTERSECTION_TIMING_CALLABLE, debug: bool = False
    ) -> None:
        """Create a sensor motion manager for queueing intersection events.

        Args:
            intersection_model (INTERSECTION_TIMING_CALLABLE): The odel to calculate
                intersections.
            debug (bool, optional): Allow debug logging to _debug_log. Defaults to False.
        """
        super().__init__()
        self._sensors: dict[SensorType, tuple[str, str]] = {}
        self._movers: dict[Actor, tuple[float, LOC_LIST, float]] = {}
        self._events: dict[Actor, list[tuple[SensorType, Process]]] = {}
        self._in_view: dict[Actor, set[SensorType]] = {}
        self._debug: bool = debug
        self._debug_data: dict[Actor, list[Any]] = {}
        self._debug_log: list[Any] = []
        self.intersection = intersection_model

    def _test_detect(self, mover: Actor) -> str | None:
        detect_state = mover._get_detection_state()
        return detect_state

    def _stop_mover(self, mover: Actor, from_not_detectable: bool = False) -> None:
        """Stop a mover.

        Args:
            mover (Actor): The moving actor.
            from_not_detectable (bool, optional): Is this was called from detectability state.
                Defaults to False.
        """
        detect_state = self._test_detect(mover)
        if detect_state is None:
            return None

        detectable: bool = getattr(mover, detect_state)
        if mover not in self._movers and not detectable:
            return None

        # Call this when a mover stops its motion
        if mover not in self._movers and not from_not_detectable:
            raise MotionAndDetectionError(f"Mover {mover} wasn't moving yet")
        elif mover not in self._movers:
            return None
        # It's possible for a mover to have no events when it stops
        # since it has no intersections. But we need the mover to exist
        # in case a new sensor pops up
        if mover in self._events:
            for _, proc in self._events.get(mover, []):
                if proc.is_alive:
                    proc.interrupt()
            del self._events[mover]

        # clear the mover references
        del self._movers[mover]
        return None

    def _mover_not_detectable(self, mover: Actor) -> None:
        """Called via DetectabilityState when a mover becomes undectable.

        Could be called for any reason; use this feature to alert sensors that
        a mover should no longer be considered by that sensor.

        Args:
            mover (Actor): The mover.
        """
        if mover in self._in_view:
            for sensor in self._in_view[mover]:
                # This will cause some old data to stick around, but that's
                # instead of making new events to clear it out and then having
                # to end those clearing events if this happens
                sensor.entity_exited_range(mover)
            del self._in_view[mover]
        # It is unsure if the user will stop the motion via a task first
        # or change detectability first
        self._stop_mover(mover, from_not_detectable=True)

    def _mover_became_detectable(self, mover: Actor) -> None:
        """Called via DetectabilityState when a mover becomes detectable.

        The actor calls this in its movement states.

        Args:
            mover (Actor): The mover.
        """
        # Before a mover is 'restarted' when becoming detectable,
        # we have to know if it's still moving
        move_states = (
            GeodeticLocationChangingState,
            CartesianLocationChangingState,
        )
        # find if there is a location changing state that is active
        locations = [
            name
            for name in mover._active_states
            if isinstance(mover._state_defs[name], move_states)
        ]
        if locations:
            msg = (
                "Setting DetectabilityState to True while "
                "locations states are active won't affect the"
                "SensorMotionManager."
            )
            warn(msg, UserWarning)

    # TODO: remove sensor or 'not active'?

    def _process_mover_sensor_pair(
        self, mover: Actor, sensor: SensorType
    ) -> list[tuple[tuple[str, float, LOC_TYPES], tuple[str, float, LOC_TYPES]]]:
        """Find the intersections (if any) b/w mover and sensor.

        Args:
            mover (Actor): The mover
            sensor (SensorType): The sensor

        Returns:
            list[tuple[str, float]]: What the movement events are are and their times
        """
        # Get pairs of "Inside/entering - Leaving/staying" to send
        # to the probability model and scheduler
        speed, waypoints, start_time = self._movers[mover]
        location_name, radius_name = self._sensors[sensor]
        location: LOC_TYPES = getattr(sensor, location_name)
        radius: float = getattr(sensor, radius_name)

        # Since waypoints connect, don't keep 'finish_in' unless
        # it's the last point in the series
        elapsed_time: float = 0.0
        inter_data: list[tuple[str, float, LOC_TYPES]] = []
        for i in range(len(waypoints) - 1):
            start, finish = waypoints[i : i + 2]
            # These times are relative to the start of the path
            intersections, times, types, path_time = self.intersection(
                start,
                finish,
                speed,
                location,
                radius,
            )

            for (
                inter,
                t,
                typ,
            ) in zip(intersections, times, types):
                if inter_data and inter_data[-1][0] == "END_INSIDE":
                    # drop that one since we are continuing on from that point
                    # and ignore this current one since it's inside already
                    if typ != "START_INSIDE":
                        raise SimulationError("START_INSIDE must follow END_INSIDE")

                    inter_data.pop()
                    continue
                inter_data.append((typ, start_time + elapsed_time + t, inter))
            elapsed_time += path_time

        # pair off the in/out
        if len(inter_data) % 2 != 0:
            raise SimulationError(f"Intersections should pair in/out or in/in, found: {inter_data}")
        pairs = [(inter_data[i], inter_data[i + 1]) for i in range(0, len(inter_data), 2)]
        if not all((a[0], b[0]) in VALID for a, b in pairs):
            raise SimulationError(f"Bad pairing of intersections: {pairs}")
        return pairs

    def _add_to_view(self, mover: Actor, sensor: SensorType) -> bool:
        """Add a mover to a sensor's view.

        Args:
            mover (Actor): Mover
            sensor (Actor): Sensor

        Returns:
            bool: If it was already in view.
        """
        if mover not in self._in_view:
            self._in_view[mover] = set()
        was_in = sensor in self._in_view[mover]
        self._in_view[mover].add(sensor)
        return was_in

    def _remove_from_view(self, mover: Actor, sensor: SensorType) -> None:
        """Remove a mover from a sensor's view.

        Args:
            mover (Actor): Mover
            sensor (Actor): Sensor
        """
        if mover not in self._in_view:
            raise MotionAndDetectionError(f"{mover} isn't in view of anything to remove.")
        if sensor not in self._in_view[mover]:
            raise MotionAndDetectionError(f"{mover} isn't in view of {sensor} to allow clearing.")
        self._in_view[mover].remove(sensor)

    def _end_notify(self, mover: Actor, sensor: SensorType, event: str) -> None:
        """End notification and give a reason.

        Args:
            mover (Actor): Mover
            sensor (Actor): Sensor
            event (str): Reason
        """
        if self._debug:
            msg = {
                "time": self.env.now,
                "event": f"Detection of a mover cancelled {event}",
                "mover": mover,
                "sensor": sensor,
            }
            self._debug_log.append(msg)

    def _notify(
        self,
        mover: Actor,
        sensor: SensorType,
        first_time: float,
        second_time: float,
        first_kind: str,
        second_kind: str,
    ) -> Generator[SimpyEvent, Any, None]:
        """Notify a sensor about a mover.

        Handles entry to exit in one go, making interrupts easier.

        Args:
            mover (Actor): Mover
            sensor (Actor): Sensor
            first_time (float): Time of the first notification
            second_time (float): Time of the second
            first_kind (str): Kind of the first
            second_kind (str): Kind of the second
        """
        # times are absolute on input to this method
        notify_time_from_now = first_time - self.env.now

        if first_kind == "START_INSIDE" or notify_time_from_now <= 0:
            was_in = self._add_to_view(mover, sensor)
            if not was_in:
                sensor.entity_entered_range(mover)
        else:
            assert first_kind == "ENTER"
            try:
                yield self.env.timeout(notify_time_from_now)
            except Interrupt:
                self._end_notify(mover, sensor, "before entry")
                return None

            sensor.entity_entered_range(mover)
            self._add_to_view(mover, sensor)

        if second_kind != "EXIT":
            return None

        end_time_from_now = second_time - self.env.now
        if end_time_from_now <= 0:
            raise MotionAndDetectionError("Detection end time is less than detection start")

        try:
            yield self.env.timeout(end_time_from_now)
        except Interrupt:
            self._end_notify(mover, sensor, "before exit")
            return None

        sensor.entity_exited_range(mover)
        self._remove_from_view(mover, sensor)
        return None

    def _schedule(
        self,
        mover: Actor,
        sensor: SensorType,
        events: tuple[tuple[str, float, LOC_TYPES], tuple[str, float, LOC_TYPES]],
    ) -> None:
        """Schedule events based on entry/exit into sensor range.

        Args:
            mover (Actor): The mover
            sensor (SensorType): The sensor
            events (list[tuple[str, float]]): Crossing events (ENTER and EXIT)
        """
        if not events:
            return
        first_kind, first_time, first_loc = events[0]
        second_kind: str = ""
        second_time: float = first_time
        second_loc: LOC_TYPES | None = None
        if len(events) > 1:
            second_kind, second_time, second_loc = events[1]

        # If both times are in the past, then the segment has already occurred
        # and we can skip it.
        if first_time <= self.env.now and second_time < self.env.now:
            return

        if self._debug:
            self._debug_data[mover].append(
                (
                    sensor,
                    [first_kind, second_kind],
                    [first_time, second_time],
                    [first_loc, second_loc],
                )
            )
            msg = {
                "time": self.env.now,
                "event": "Scheduling sensor detecting mover",
                "mover": mover,
                "sensor": sensor,
            }
            self._debug_log.append(msg)

        proc = self.env.process(
            self._notify(mover, sensor, first_time, second_time, first_kind, second_kind)
        )
        if mover not in self._events:
            self._events[mover] = [
                (sensor, proc),
            ]
        else:
            self._events[mover].append((sensor, proc))

    def _find_intersections(
        self,
        mover_list: list[Actor] | None = None,
        sensor_list: list[SensorType] | None = None,
    ) -> None:
        """Find all paired intersections and schedule them.

        Optionally, use a reduced list of either movers or sensors.

        Args:
            mover_list (list[Actor] | None, optional): Movers to consider.
                Defaults to None (all movers).
            sensor_list (list[SensorType] | None, optional): Sensors to consider.
                Defaults to None (all sensors).
        """
        movers = list(self._movers.keys()) if mover_list is None else mover_list
        sensors = list(self._sensors.keys()) if sensor_list is None else sensor_list
        for m in movers:
            for s in sensors:
                inter_pairs = self._process_mover_sensor_pair(m, s)
                for pair in inter_pairs:
                    self._schedule(m, s, pair)

    def _start_mover(self, mover: Actor, speed: float, waypoints: LOC_LIST) -> None:
        """Start a mover's motion and find intersections with sensors.

        Args:
            mover (Actor): The mover
            speed (float): Speed (in model units)
            waypoints (LOC_LIST): Waypoint of travel.
        """
        detect_state = self._test_detect(mover)
        if detect_state is None:
            return

        detectable = getattr(mover, detect_state)
        # Since this class examines self._movers when mover stops, we need to put in
        # some data about that so we get the right errors if new motion starts
        # when this motion hasn't ended.
        if not detectable:
            self._movers[mover] = (0.0, [], 0.0)
            return None

        if mover in self._movers:
            raise MotionAndDetectionError(
                f"Mover: {mover} is already known to be on a path. Did you forget to stop it?"
            )

        if self._debug and mover not in self._debug_data:
            self._debug_data[mover] = []
        # TODO: Waypoints need to start with the movers current location
        self._movers[mover] = (speed, waypoints, self.env.now)
        self._find_intersections(mover_list=[mover], sensor_list=None)
        return None

    def add_sensor(
        self,
        sensor: SensorType,
        location_attr_name: str = "location",
        radius_attr_name: str = "radius",
    ) -> None:
        """Add a sensor to the motion manager.

        Args:
            sensor (SensorType): The sensor object
            location_attr_name (str, optional): Name of the location attribute.
                Defaults to "location".
            radius_attr_name (str, optional): Name of the radius attribute. Defaults to "radius".
        """
        # test the sensor for earlier errors about improperly-defined methods
        required_methods = ["entity_entered_range", "entity_exited_range"]
        for req in required_methods:
            if not hasattr(sensor, req):
                raise NotImplementedError(f"Sensor {sensor} does not have '{req}' method!")
        for attr in [location_attr_name, radius_attr_name]:
            _attr = getattr(sensor, attr, None)
            if _attr is None:
                raise SimulationError(f"Sensor {sensor} has no attribute: {attr}")

        self._sensors[sensor] = (location_attr_name, radius_attr_name)
        self._find_intersections(mover_list=None, sensor_list=[sensor])



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from dataclasses import FrozenInstanceError
from math import radians, sqrt

import pytest

import upstage_des.api as UP
from upstage_des.geography import Spherical, get_intersection_locations

STAGE_SETUP = dict(
    altitude_units="ft",
    distance_units="nmi",
    intersection_model=get_intersection_locations,
    stage_model=Spherical,
)


def test_basics() -> None:
    with UP.EnvironmentContext():
        with pytest.raises(NotImplementedError):
            l1 = UP.Location()
            l1 - 3


def test_hashable() -> None:
    with UP.EnvironmentContext():
        p1 = [10, 10]
        point_1 = UP.CartesianLocation(*p1)
        key = point_1._key()
        assert key == (10, 10, 0.0, False)
        # assert getattr(point_1, "__hash__") is not None
        _ = {point_1: 1.0}


def test_cartesian() -> None:
    with UP.EnvironmentContext():
        for k, v in STAGE_SETUP.items():
            UP.add_stage_variable(k, v)

        p1 = [10.0, 10.0]
        p2 = [1.0, 2.0, 3.0]
        origin = UP.CartesianLocation(0, 0)
        point_1 = UP.CartesianLocation(*p1)
        point_2 = UP.CartesianLocation(*p2)

        assert origin - point_1 == sqrt(200)
        assert origin - point_1 == point_1 - origin

        assert point_1 - point_2 > 0
        for index, value in enumerate(p2):
            assert point_2[index] == value

        assert "10" in point_1.__repr__()

        with pytest.raises(ValueError):
            point_2[3] is None

        with pytest.raises(ValueError):
            point_1 - 10

        with pytest.raises(ValueError):
            point_2 == 10

        point_3 = UP.CartesianLocation(*p2, use_altitude_units=True)

        assert point_3 != point_2

        point_2a = UP.CartesianLocation(1, 2, 3.000000001)
        assert point_2 == point_2a, f"Nearly equal points are still {point_2 - point_2a} too far"


def test_geodetic() -> None:
    with UP.EnvironmentContext():
        for k, v in STAGE_SETUP.items():
            UP.add_stage_variable(k, v)

        lat, lon, alt = 33, -86, 1000
        loc_up = UP.GeodeticLocation(lat, lon, alt)

        with pytest.raises(FrozenInstanceError):
            loc_up.alt = 5_000

        loc_up_rad = loc_up.to_radians()
        assert loc_up_rad.lat == radians(loc_up.lat)
        assert loc_up_rad.lon == radians(loc_up.lon)
        assert loc_up_rad.alt == loc_up.alt

        loc_up_rad_1 = UP.GeodeticLocation(
            lat=radians(loc_up.lat),
            lon=radians(loc_up.lon),
            in_radians=True,
        )
        assert loc_up_rad_1.lat == loc_up_rad.lat
        assert loc_up_rad_1.lon == loc_up_rad.lon

        assert loc_up_rad == loc_up_rad.to_radians()
        assert loc_up == loc_up.to_degrees()

        assert loc_up_rad == loc_up
        assert loc_up == loc_up_rad

        assert loc_up_rad - UP.GeodeticLocation(10, 10) > 0
        assert UP.GeodeticLocation(10, 10) - loc_up_rad > 0


def test_data_objects() -> None:
    cart1 = UP.CartesianLocationData(1.0, 2.1, 3.2)
    cart2 = UP.CartesianLocationData(1.0, 2.1, 3.2)
    assert cart1 == cart2

    with pytest.raises(ValueError):
        cart1 == (1.0, 2.1, 3.2)

    geo1 = UP.GeodeticLocationData(13.0, 12.1, 11.2)
    geo2 = UP.GeodeticLocationData(13.0, 12.1, 11.2)
    assert geo1 == geo2

    geo3 = UP.GeodeticLocationData(radians(13.0), radians(12.1), 11.2, in_radians=True)
    assert geo1 == geo3
    assert geo3 == geo1

    with pytest.raises(ValueError):
        geo1 == (13.0, 12.1, 11.2)

    with UP.EnvironmentContext():
        for k, v in STAGE_SETUP.items():
            UP.add_stage_variable(k, v)

        loc1 = geo1.make_location()
        assert isinstance(loc1, UP.GeodeticLocation)

        loc3 = geo3.make_location()

        assert loc1 - loc3 == 0.0
        assert loc1 == loc3

        loc1 = cart1.make_location()
        loc2 = cart2.make_location()
        assert loc1 == loc2



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Help for typing task and simpy generators."""

from collections.abc import Generator
from typing import Any

from simpy import Event as SimEvent

from upstage_des.events import BaseEvent

TASK_GEN = Generator[BaseEvent, Any, None]
SIMPY_GEN = Generator[SimEvent, Any, Any]



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest

import upstage_des.api as UP
from upstage_des.type_help import TASK_GEN
from upstage_des.utils import waypoint_time_and_dist


class Plane(UP.Actor):
    speed = UP.State[float]()
    location = UP.CartesianLocationChangingState()
    fuel = UP.LinearChangingState()
    fuel_burn = UP.State[float]()


class Fly(UP.Task):
    def task(self, *, actor: Plane) -> TASK_GEN:
        fly_to: list[UP.CartesianLocation] = self.get_actor_knowledge(
            actor, "destination", must_exist=True
        )
        time, dist = waypoint_time_and_dist(actor.location, fly_to, actor.speed)
        print(f"Rehearsing the task: {self._rehearsing}")
        print(f"\tFlying {dist:.2f} units over {time:.2f} hrs")
        actor.activate_linear_state(
            state="fuel",
            rate=-actor.fuel_burn,
            task=self,
        )
        actor.activate_location_state(
            state="location",
            speed=actor.speed,
            waypoints=fly_to,
            task=self,
        )
        yield UP.Wait(time)
        actor.deactivate_all_states(task=self)


class Search(UP.Task):
    def task(self, *, actor: Plane) -> TASK_GEN:
        search_event = actor.create_knowledge_event(
            name="FOUND SURVIVOR",
            rehearsal_time_to_complete=0.5,
        )
        actor.activate_linear_state(
            state="fuel",
            rate=-actor.fuel_burn,
            task=self,
        )
        yield search_event
        actor.deactivate_all_states(task=self)


class Land(UP.Task):
    def task(self, *, actor: Plane) -> TASK_GEN:
        # Do a landing of some kind
        event = actor.create_knowledge_event(name="DONE", rehearsal_time_to_complete=10.0)
        yield event


def some_preference_function(
    spots: list[UP.CartesianLocation],
) -> UP.CartesianLocation | None:
    """Choose a spot. bvv z

    Args:
        spots (list[UP.CartesianLocation]): List of spots to search at

    Returns:
        UP.CartesianLocation | None: A spot to search or None if we should go home
    """
    return spots[0]


class Planner(UP.DecisionTask):
    def make_decision(self, *, actor: Plane) -> None:
        go_to_loc = some_preference_function(self.stage.search_spots)
        if go_to_loc is None:  # implies we are done with searching
            self.set_actor_task_queue(actor, ["Fly", "Land"])
        else:
            self.set_actor_knowledge(actor, "destination", go_to_loc, overwrite=True)
            self.set_actor_task_queue(actor, ["Fly", "Search"])

    def rehearse_decision(self, *, actor: Plane) -> None:
        # Pop off a destination from the queue, or go "home"
        next_dests: list[list[UP.CartesianLocation]] | None = self.get_actor_knowledge(
            actor, "destination_plan", must_exist=False
        )
        dests: list[UP.CartesianLocation]
        task_queue: list[str]
        if not next_dests:  # fly home
            dests = [UP.CartesianLocation(0, 0)]
            task_queue = ["Fly", "Land"]
        else:  # pop a location from the plan
            dests = next_dests.pop(0)
            self.set_actor_knowledge(actor, "destination_plan", next_dests, overwrite=True)
            task_queue = ["Fly", "Search"]

        self.set_actor_knowledge(actor, "destination", dests, overwrite=True)
        self.set_actor_task_queue(actor, task_queue)


task_classes = {"Fly": Fly, "Search": Search, "Planner": Planner, "Land": Land}
task_links = {
    "Fly": UP.TaskLinks(default="Search", allowed=["Fly", "Land", "Search"]),
    "Search": UP.TaskLinks(default="Planner", allowed=["Planner"]),
    "Planner": UP.TaskLinks(default="Fly", allowed=["Fly"]),
    "Land": UP.TaskLinks(default=None, allowed=["Fly"]),
}
search_network = UP.TaskNetworkFactory("SearchNet", task_classes, task_links)


def test_model() -> None:
    with UP.EnvironmentContext() as env:
        search_locs = [
            [UP.CartesianLocation(x, y)]
            for x, y in [
                (10, 20),
                (30, 10),
                (15, 15),
            ]
        ]

        plane = Plane(
            name="searcher",
            speed=2,
            fuel=200,
            fuel_burn=5.0,
            location=UP.CartesianLocation(20, 10),
            debug_log=True,
        )
        net = search_network.make_network()
        plane.add_task_network(net)

        new_plane = plane.rehearse_network(
            net.name,
            task_name_list=["Planner", "Fly", "Search"],
            knowledge={"destination_plan": search_locs},
            end_task="Land",
        )
        print(f"Fuel left: {new_plane.fuel}")
        print(f"Time passed: {new_plane.env.now}")
        print(f"Actual time passed: {env.now}")
        assert pytest.approx(new_plane.fuel) == 6.18148
        assert pytest.approx(new_plane.env.now) == 38.76370356758358
        assert plane.fuel == 200
        assert env.now == 0



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from typing import Any

import simpy as SIM

import upstage_des.api as UP
from upstage_des.type_help import TASK_GEN

from .mover import Mover


class Mothership(Mover):
    fuel_ports_in_use = UP.State[int](valid_types=(int,))
    fuel_ports_max = UP.State[int](valid_types=(int,), frozen=True)
    messages = UP.ResourceState[UP.SelfMonitoringStore](
        default=UP.SelfMonitoringStore, valid_types=(UP.SelfMonitoringStore, SIM.Store)
    )


class DispenseFuel(UP.Task):
    def task(self, *, actor: Mothership) -> TASK_GEN:
        draws = self.get_actor_knowledge(actor, "fuel_users", must_exist=False)
        draws = {} if draws is None else draws
        total_draw = sum(draws.values())
        actor.activate_state(
            state="fuel",
            task=self,
            rate=total_draw,
        )
        # Infinite wait!
        # We use Nucleus to go to the interrupt.
        yield UP.Event()

    def on_interrupt(self, *, actor: Mothership, cause: Any) -> UP.InterruptStates:
        # No matter what, restart
        return self.INTERRUPT.RESTART


give_fuel_factory = UP.TaskNetworkFactory.from_single_looping("GiveFuel", DispenseFuel)


class CrewMember(UP.Task):
    def _user_add(self, actor: Mothership, vehicle: Mover, add: float) -> int:
        know = self.get_actor_knowledge(actor, "fuel_users")
        know = {} if know is None else know
        know[vehicle] = add
        self.set_actor_knowledge(actor, "fuel_users", know, overwrite=True)
        return len(know)

    def _user_remove(self, actor: Mothership, vehicle: Mover) -> int:
        know = self.get_actor_knowledge(actor, "fuel_users")
        know = {} if know is None else know
        del know[vehicle]
        self.set_actor_knowledge(actor, "fuel_users", know, overwrite=True)
        return len(know)

    def task(self, *, actor: Mothership) -> TASK_GEN:
        # receive a message that someone is ready, then update fuel_users
        msg = yield UP.Get(actor.messages)
        vehicle, draw = msg
        if draw == 0:
            ports = self._user_remove(actor, vehicle)
        else:
            ports = self._user_add(actor, vehicle, draw)
            # kill some time to send a message back
            yield UP.Wait(5 / 60)
            yield UP.Put(vehicle.messages, "GO")
        actor.fuel_ports_in_use = ports


crew_factory = UP.TaskNetworkFactory.from_single_looping("crew", CrewMember)



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Comms message and commander classes."""

from collections.abc import Generator
from dataclasses import dataclass
from typing import Any
from uuid import uuid4

from simpy import Event as SimpyEvent
from simpy import Store

from upstage_des.actor import Actor
from upstage_des.base import ENV_CONTEXT_VAR, SimulationError, UpstageBase
from upstage_des.events import Put
from upstage_des.states import CommunicationStore
from upstage_des.task import process


@dataclass
class MessageContent:
    """Message content data object."""

    data: dict
    message: str | None = None


@dataclass
class Message:
    """A message data object."""

    sender: Actor
    content: MessageContent
    destination: Actor

    header: str | None = None
    time_sent: float | None = None
    time_received: float | None = None

    def __post_init__(self) -> None:
        self.uid = uuid4()
        self.time_created = ENV_CONTEXT_VAR.get().now

    def __hash__(self) -> int:
        return hash(self.uid)


class CommsManager(UpstageBase):
    """A class to manage point to point transfer of communications.

    Works through simpy.Store or similar interfaces. Allows for degraded comms and comms retry.

    If an Actor contains a `CommunicationStore`, this object will detect that
    and use it as a destination. In that case, you also do not need to connect
    the actor to this object.

    Example:
        >>> class Talker(UP.Actor):
        >>>     comms = UP.ResourceState[SIM.Store](default=SIM.Store)
        >>>
        >>> talker1 = Talker(name='MacReady')
        >>> talker2 = Talker(name='Childs')
        >>>
        >>> comm_station = UP.CommsManager(name="Outpost 31", mode="voice")
        >>> comm_station.connect(talker1, talker1.comms)
        >>> comm_station.connect(talker2, talker2.comms)
        >>>
        >>> comm_station.run()
        >>>
        >>> # Typically, do this inside a task or somewhere else
        >>> putter = comm_station.make_put(
        >>>     message="Grab your flamethrower!",
        >>>     source=talker1,
        >>>     destination=talker2,
        >>>     rehearsal_time_to_complete=0.0,
        >>> )
        >>> yield putter
        ...
        >>> env.run()
        >>> talker2.comms.items
            [Message(sender=Talker: MacReady, message='Grab your flamethrower!',
            destination=Talker: Childs)]
    """

    def __init__(
        self,
        *,
        name: str,
        mode: str | None = None,
        init_entities: list[tuple[Actor, str]] | None = None,
        send_time: float = 0.0,
        retry_max_time: float = 1.0,
        retry_rate: float = 0.166667,
        debug_logging: bool = False,
    ) -> None:
        """Create a comms transfer manager.

        Parameters
        ----------
        name : str
            Give the instance a unique name for logging purposes
        mode: str
            The name of the mode comms are occurring over. Used for automated
            detection of actor comms interfaces.
            Default is None, which requires explicit connections.
        init_entities : List[Tuple(instance, str)], optional
            Entities who have a comms store to let the manager know about. The
            tuples are (entity_instance, entity's comms input store's name), by default None
        send_time : float, optional
            Time to send a message, by default 0.0
        retry_max_time : float, optional
            Amount of time (in sim units) to try resending a message, by default 1
        retry_rate : float, optional
            How often (in sim units) to try re-sending a message, by default 10/60
        debug_logging : bool, optional
            Turn on or off logging, by default False
        """
        super().__init__()
        self.name = name
        self.mode = mode
        self.comms_degraded: bool = False
        self.retry_max_time = retry_max_time
        self.retry_rate = retry_rate
        self.send_time = send_time
        self.incoming = Store(env=self.env)
        self.connected: dict[Actor, str] = {}
        self.blocked_links: list[tuple[Actor, Actor]] = []
        if init_entities is not None:
            for entity, comms_store_name in init_entities:
                self.connect(entity, comms_store_name)
        self.debug_log: list[str | dict] = []
        self.debug_logging: bool = debug_logging

    @staticmethod
    def clean_message(message: str | Message) -> MessageContent:
        """Test to see if an object is a message.

        If it is, return the message contents only. Otherwise return the message.

        Args:
            message (str | Message): The message to clean

        Returns:
            MessageContent: The message as a message content object.
        """
        if isinstance(message, Message):
            return message.content
        return MessageContent(data={"message": message})

    def connect(self, entity: Actor, comms_store_name: str) -> None:
        """Connect an actor and its comms store to this comms manager.

        Args:
            entity (Actor): The actor that will send/receive.
            comms_store_name (str): The store state name for receiving
        """
        self.connected[entity] = comms_store_name

    def store_from_actor(self, actor: Actor) -> Store:
        """Retrieve a communications store from an actor.

        Args:
            actor (Actor): The actor

        Returns:
            Store: A Comms store.
        """
        if actor not in self.connected:
            try:
                msg_store_name = actor._get_matching_state(CommunicationStore, {"_mode": self.mode})
            except SimulationError as e:
                e.add_note(f"No comms destination on actor {actor}")
                raise e
        else:
            msg_store_name = self.connected[actor]

        if msg_store_name is None:
            raise SimulationError(f"No comms store on {actor}")
        store: Store | None = getattr(actor, msg_store_name)
        if store is None:
            raise SimulationError(f"Bad comms store name: {msg_store_name} on {actor}")
        return store

    def make_put(
        self,
        message: str | Message | MessageContent | dict,
        source: Actor,
        destination: Actor,
        rehearsal_time_to_complete: float = 0.0,
    ) -> Put:
        """Create a Put request for a message into the CommsManager.

        Parameters
        ----------
        source :
            The message sender
        destination :
            The message receiver, who must be connected to the CommsManager
        message :
            Arbitrary data to send
        rehearsal_time_to_complete : float, optional
            Planning time to complete the event (see Put), by default 0.0

        Returns:
        -------
        Put
            UPSTAGE Put event object to yield from a task
        """
        use: Message
        if isinstance(message, Message):
            use = message
        elif isinstance(message, MessageContent):
            use = Message(sender=source, content=message, destination=destination)
        else:
            content = (
                MessageContent(data=message)
                if isinstance(message, dict)
                else MessageContent(data={}, message=message)
            )
            use = Message(sender=source, content=content, destination=destination)

        return Put(
            self.incoming,
            use,
            rehearsal_time_to_complete=rehearsal_time_to_complete,
        )

    @process
    def _do_transmit(
        self, message: Message, destination: Actor
    ) -> Generator[SimpyEvent, None, None]:
        start_time = self.env.now
        while self.comms_degraded or self.test_if_link_is_blocked(message):
            if self.debug_logging:
                msg = {
                    "time": self.env.now,
                    "event": "Can't sent, waiting",
                    "message": message,
                    "destination": destination,
                }
                self.debug_log.append(msg)

            elapsed_time = self.env.now - start_time
            if elapsed_time > self.retry_max_time:
                if self.debug_logging:
                    msg = {
                        "time": self.env.now,
                        "event": "Stopped trying to send",
                        "message": message,
                        "destination": destination,
                    }
                    self.debug_log.append(msg)
                return

            yield self.env.timeout(self.retry_rate)

        if self.send_time > 0:
            yield self.env.timeout(self.send_time)

        if self.debug_logging:
            msg = {
                "time": self.env.now,
                "event": "Sent message",
                "message": message,
                "destination": destination,
            }
            self.debug_log.append(msg)

        # update the send time
        message.time_sent = self.env.now
        store = self.store_from_actor(destination)
        yield store.put(message)

    @process
    def run(self) -> Generator[SimpyEvent, Any, None]:
        """Run the communications message passing.

        Yields:
            Generator[SimpyEvent, Any, None]: Simpy Process
        """
        while True:
            message = yield self.incoming.get()
            dest = message.destination
            self._do_transmit(message, dest)

    def _link_compare(self, a_test: Actor, b_test: Actor) -> bool:
        # python fails at comparing existence and tries a different equality test
        for a, b in self.blocked_links:
            if a_test is a and b_test is b:
                return True
        return False

    def test_if_link_is_blocked(self, message: Message) -> bool:
        """Test if a link is blocked.

        Args:
            message (Message): Message with sender/destination data.

        Returns:
            bool: If the link is blocked.
        """
        if self._link_compare(message.sender, message.destination):
            return True
        return False
